# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Cancel the redundant (self-adjoint) gates through commutation relations."""
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.circuit.commutation_library import StandardGateCommutations

from qiskit.circuit.library.standard_gates.u1 import U1Gate
from qiskit.circuit.library.standard_gates.p import PhaseGate
from qiskit.circuit.library.standard_gates.rz import RZGate
from qiskit._accelerate import commutation_cancellation
from qiskit._accelerate.commutation_checker import CommutationChecker

from qiskit.transpiler.passes.utils.control_flow import trivial_recurse

_CUTOFF_PRECISION = 1e-5


class CommutativeCancellation(TransformationPass):
    """Cancel the redundant (self-adjoint) gates through commutation relations.

    Pass for cancelling self-inverse gates/rotations. The cancellation utilizes
    the commutation relations in the circuit. Gates considered include::

        H, X, Y, Z, CX, CY, CZ
    """

    def __init__(self, basis_gates=None, target=None):
        """
        CommutativeCancellation initializer.

        Args:
            basis_gates (list[str]): Basis gates to consider, e.g.
                ``['u3', 'cx']``. For the effects of this pass, the basis is
                the set intersection between the ``basis_gates`` parameter
                and the gates in the dag.
            target (Target): The :class:`~.Target` representing the target backend, if both
                ``basis_gates`` and ``target`` are specified then this argument will take
                precedence and ``basis_gates`` will be ignored.
        """
        super().__init__()
        if basis_gates:
            self.basis = set(basis_gates)
        else:
            self.basis = set()
        self.target = target
        if target is not None:
            self.basis = set(target.operation_names)

        self._var_z_map = {"rz": RZGate, "p": PhaseGate, "u1": U1Gate}

        self._z_rotations = {"p", "z", "u1", "rz", "t", "s"}
        self._x_rotations = {"x", "rx"}
        self._gates = {"cx", "cy", "cz", "h", "y"}  # Now the gates supported are hard-coded

        # build a commutation checker restricted to the gates we cancel -- the others we
        # do not have to investigate, which allows to save time
        self._commutation_checker = CommutationChecker(
            StandardGateCommutations, gates=self._gates | self._z_rotations | self._x_rotations
        )

    @trivial_recurse
    def run(self, dag):
        """Run the CommutativeCancellation pass on `dag`.

        Args:
            dag (DAGCircuit): the DAG to be optimized.

        Returns:
            DAGCircuit: the optimized DAG.
        """
        commutation_cancellation.cancel_commutations(dag, self._commutation_checker, self.basis)
        return dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import copy
import itertools
import pickle
import random
import unittest

import ddt
import numpy as np

from qiskit import transpile
from qiskit.circuit import Measure, Parameter, library, QuantumCircuit
from qiskit.exceptions import QiskitError
from qiskit.quantum_info import SparseObservable, SparsePauliOp, Pauli, PauliList
from qiskit.transpiler import Target

from test import QiskitTestCase, combine  # pylint: disable=wrong-import-order


def single_cases():
    return [
        SparseObservable.zero(0),
        SparseObservable.zero(10),
        SparseObservable.identity(0),
        SparseObservable.identity(1_000),
        SparseObservable.from_label("IIXIZI"),
        SparseObservable.from_list([("YIXZII", -0.25), ("01rl+-", 0.25 + 0.5j)]),
        # Includes a duplicate entry.
        SparseObservable.from_list([("IXZ", -0.25), ("01I", 0.25 + 0.5j), ("IXZ", 0.75)]),
    ]


def lnn_target(num_qubits):
    """Create a simple `Target` object with an arbitrary basis-gate set, and open-path
    connectivity."""
    out = Target()
    out.add_instruction(library.RZGate(Parameter("a")), {(q,): None for q in range(num_qubits)})
    out.add_instruction(library.SXGate(), {(q,): None for q in range(num_qubits)})
    out.add_instruction(Measure(), {(q,): None for q in range(num_qubits)})
    out.add_instruction(
        library.CXGate(),
        {
            pair: None
            for lower in range(num_qubits - 1)
            for pair in [(lower, lower + 1), (lower + 1, lower)]
        },
    )
    return out


class AllowRightArithmetic:
    """Some type that implements only the right-hand-sided arithmatic operations, and allows
    `SparseObservable` to pass through them.

    The purpose of this is to detect that `SparseObservable` is correctly delegating binary
    operators to the other type if given an object it cannot coerce because of its type."""

    SENTINEL = object()

    __radd__ = __rsub__ = __rmul__ = __rtruediv__ = __rxor__ = lambda self, other: self.SENTINEL


@ddt.ddt
class TestSparseObservable(QiskitTestCase):
    def test_default_constructor_pauli(self):
        data = Pauli("IXYIZ")
        self.assertEqual(SparseObservable(data), SparseObservable.from_pauli(data))
        self.assertEqual(
            SparseObservable(data, num_qubits=data.num_qubits), SparseObservable.from_pauli(data)
        )
        with self.assertRaisesRegex(ValueError, "explicitly given 'num_qubits'"):
            SparseObservable(data, num_qubits=data.num_qubits + 1)

        with_phase = Pauli("-jIYYXY")
        self.assertEqual(SparseObservable(with_phase), SparseObservable.from_pauli(with_phase))
        self.assertEqual(
            SparseObservable(with_phase, num_qubits=data.num_qubits),
            SparseObservable.from_pauli(with_phase),
        )

        self.assertEqual(SparseObservable(Pauli("")), SparseObservable.from_pauli(Pauli("")))

    def test_default_constructor_sparse_pauli_op(self):
        data = SparsePauliOp.from_list([("IIXIY", 1.0), ("XYYZI", -0.25), ("XYIYY", -0.25 + 0.75j)])
        self.assertEqual(SparseObservable(data), SparseObservable.from_sparse_pauli_op(data))
        self.assertEqual(
            SparseObservable(data, num_qubits=data.num_qubits),
            SparseObservable.from_sparse_pauli_op(data),
        )
        with self.assertRaisesRegex(ValueError, "explicitly given 'num_qubits'"):
            SparseObservable(data, num_qubits=data.num_qubits + 1)
        with self.assertRaisesRegex(TypeError, "complex-typed coefficients"):
            SparseObservable(SparsePauliOp(["XX"], [Parameter("x")]))

    def test_default_constructor_label(self):
        data = "IIXI+-I01rlIYZ"
        self.assertEqual(SparseObservable(data), SparseObservable.from_label(data))
        self.assertEqual(
            SparseObservable(data, num_qubits=len(data)), SparseObservable.from_label(data)
        )
        with self.assertRaisesRegex(ValueError, "explicitly given 'num_qubits'"):
            SparseObservable(data, num_qubits=len(data) + 1)

    def test_default_constructor_list(self):
        data = [("IXIIZ", 0.5), ("+I-II", 1.0 - 0.25j), ("IIrlI", -0.75)]
        self.assertEqual(SparseObservable(data), SparseObservable.from_list(data))
        self.assertEqual(SparseObservable(data, num_qubits=5), SparseObservable.from_list(data))
        with self.assertRaisesRegex(ValueError, "label with length 5 cannot be added"):
            SparseObservable(data, num_qubits=4)
        with self.assertRaisesRegex(ValueError, "label with length 5 cannot be added"):
            SparseObservable(data, num_qubits=6)
        self.assertEqual(
            SparseObservable([], num_qubits=5), SparseObservable.from_list([], num_qubits=5)
        )

    def test_default_constructor_sparse_list(self):
        data = [("ZX", (0, 3), 0.5), ("-+", (2, 4), 1.0 - 0.25j), ("rl", (2, 1), -0.75)]
        self.assertEqual(
            SparseObservable(data, num_qubits=5),
            SparseObservable.from_sparse_list(data, num_qubits=5),
        )
        self.assertEqual(
            SparseObservable(data, num_qubits=10),
            SparseObservable.from_sparse_list(data, num_qubits=10),
        )
        with self.assertRaisesRegex(ValueError, "'num_qubits' must be provided"):
            SparseObservable(data)
        self.assertEqual(
            SparseObservable([], num_qubits=5), SparseObservable.from_sparse_list([], num_qubits=5)
        )

    def test_default_constructor_copy(self):
        base = SparseObservable.from_list([("IXIZIY", 1.0), ("+-rl01", -1.0j)])
        copied = SparseObservable(base)
        self.assertEqual(base, copied)
        self.assertIsNot(base, copied)

        # Modifications to `copied` don't propagate back.
        copied.coeffs[1] = -0.5j
        self.assertNotEqual(base, copied)

        with self.assertRaisesRegex(ValueError, "explicitly given 'num_qubits'"):
            SparseObservable(base, num_qubits=base.num_qubits + 1)

    def test_default_constructor_term(self):
        expected = SparseObservable.from_list([("IIZXII+-", 2j)])
        self.assertEqual(SparseObservable(expected[0]), expected)

    def test_default_constructor_term_iterable(self):
        expected = SparseObservable.from_list([("IIZXII+-", 2j), ("rlIIIIII", 0.5)])
        terms = [expected[0], expected[1]]
        self.assertEqual(SparseObservable(list(terms)), expected)
        self.assertEqual(SparseObservable(tuple(terms)), expected)
        self.assertEqual(SparseObservable(term for term in terms), expected)

    def test_default_constructor_failed_inference(self):
        with self.assertRaises(TypeError):
            # Mixed dense/sparse list.
            SparseObservable([("IIXIZ", 1.0), ("+-", (2, 3), -1.0)], num_qubits=5)

    def test_num_qubits(self):
        self.assertEqual(SparseObservable.zero(0).num_qubits, 0)
        self.assertEqual(SparseObservable.zero(10).num_qubits, 10)

        self.assertEqual(SparseObservable.identity(0).num_qubits, 0)
        self.assertEqual(SparseObservable.identity(1_000_000).num_qubits, 1_000_000)

    def test_num_terms(self):
        self.assertEqual(SparseObservable.zero(0).num_terms, 0)
        self.assertEqual(SparseObservable.zero(10).num_terms, 0)
        self.assertEqual(SparseObservable.identity(0).num_terms, 1)
        self.assertEqual(SparseObservable.identity(1_000_000).num_terms, 1)
        self.assertEqual(
            SparseObservable.from_list([("IIIXIZ", 1.0), ("YY+-II", 0.5j)]).num_terms, 2
        )

    def test_len(self):
        self.assertEqual(len(SparseObservable.zero(0)), 0)
        self.assertEqual(len(SparseObservable.zero(10)), 0)
        self.assertEqual(len(SparseObservable.identity(0)), 1)
        self.assertEqual(len(SparseObservable.identity(1_000_000)), 1)
        self.assertEqual(len(SparseObservable.from_list([("IIIXIZ", 1.0), ("YY+-II", 0.5j)])), 2)

    def test_bit_term_enum(self):
        # These are very explicit tests that effectively just duplicate magic numbers, but the point
        # is that those magic numbers are required to be constant as their values are part of the
        # public interface.

        self.assertEqual(
            set(SparseObservable.BitTerm),
            {
                SparseObservable.BitTerm.X,
                SparseObservable.BitTerm.Y,
                SparseObservable.BitTerm.Z,
                SparseObservable.BitTerm.PLUS,
                SparseObservable.BitTerm.MINUS,
                SparseObservable.BitTerm.RIGHT,
                SparseObservable.BitTerm.LEFT,
                SparseObservable.BitTerm.ZERO,
                SparseObservable.BitTerm.ONE,
            },
        )
        # All the enumeration items should also be integers.
        self.assertIsInstance(SparseObservable.BitTerm.X, int)
        values = {
            "X": 0b00_10,
            "Y": 0b00_11,
            "Z": 0b00_01,
            "PLUS": 0b10_10,
            "MINUS": 0b01_10,
            "RIGHT": 0b10_11,
            "LEFT": 0b01_11,
            "ZERO": 0b10_01,
            "ONE": 0b01_01,
        }
        self.assertEqual({name: getattr(SparseObservable.BitTerm, name) for name in values}, values)

        # The single-character label aliases can be accessed with index notation.
        labels = {
            "X": SparseObservable.BitTerm.X,
            "Y": SparseObservable.BitTerm.Y,
            "Z": SparseObservable.BitTerm.Z,
            "+": SparseObservable.BitTerm.PLUS,
            "-": SparseObservable.BitTerm.MINUS,
            "r": SparseObservable.BitTerm.RIGHT,
            "l": SparseObservable.BitTerm.LEFT,
            "0": SparseObservable.BitTerm.ZERO,
            "1": SparseObservable.BitTerm.ONE,
        }
        self.assertEqual({label: SparseObservable.BitTerm[label] for label in labels}, labels)

    @ddt.idata(single_cases())
    def test_pickle(self, observable):
        self.assertEqual(observable, copy.copy(observable))
        self.assertIsNot(observable, copy.copy(observable))
        self.assertEqual(observable, copy.deepcopy(observable))
        self.assertEqual(observable, pickle.loads(pickle.dumps(observable)))

    @ddt.data(
        # This is every combination of (0, 1, many) for (terms, qubits, non-identites per term).
        SparseObservable.zero(0),
        SparseObservable.zero(1),
        SparseObservable.zero(10),
        SparseObservable.identity(0),
        SparseObservable.identity(1),
        SparseObservable.identity(1_000),
        SparseObservable.from_label("IIXIZI"),
        SparseObservable.from_label("X"),
        SparseObservable.from_list([("YIXZII", -0.25), ("01rl+-", 0.25 + 0.5j)]),
    )
    def test_repr(self, data):
        # The purpose of this is just to test that the `repr` doesn't crash, rather than asserting
        # that it has any particular form.
        self.assertIsInstance(repr(data), str)
        self.assertIn("SparseObservable", repr(data))

    def test_from_raw_parts(self):
        # Happiest path: exactly typed inputs.
        num_qubits = 100
        terms = np.full((num_qubits,), SparseObservable.BitTerm.Z, dtype=np.uint8)
        indices = np.arange(num_qubits, dtype=np.uint32)
        coeffs = np.ones((num_qubits,), dtype=complex)
        boundaries = np.arange(num_qubits + 1, dtype=np.uintp)
        observable = SparseObservable.from_raw_parts(num_qubits, coeffs, terms, indices, boundaries)
        self.assertEqual(observable.num_qubits, num_qubits)
        np.testing.assert_equal(observable.bit_terms, terms)
        np.testing.assert_equal(observable.indices, indices)
        np.testing.assert_equal(observable.coeffs, coeffs)
        np.testing.assert_equal(observable.boundaries, boundaries)

        self.assertEqual(
            observable,
            SparseObservable.from_raw_parts(
                num_qubits, coeffs, terms, indices, boundaries, check=False
            ),
        )

        # At least the initial implementation of `SparseObservable` requires `from_raw_parts` to be
        # a copy constructor in order to allow it to be resized by Rust space.  This is checking for
        # that, but if the implementation changes, it could potentially be relaxed.
        self.assertFalse(np.may_share_memory(observable.coeffs, coeffs))

        # Conversion from array-likes, including mis-typed but compatible arrays.
        observable = SparseObservable.from_raw_parts(
            num_qubits, list(coeffs), tuple(terms), observable.indices, boundaries.astype(np.int16)
        )
        self.assertEqual(observable.num_qubits, num_qubits)
        np.testing.assert_equal(observable.bit_terms, terms)
        np.testing.assert_equal(observable.indices, indices)
        np.testing.assert_equal(observable.coeffs, coeffs)
        np.testing.assert_equal(observable.boundaries, boundaries)

        # Construction of zero operator.
        self.assertEqual(
            SparseObservable.from_raw_parts(10, [], [], [], [0]), SparseObservable.zero(10)
        )

        # Construction of an operator with an intermediate identity term.  For the initial
        # constructor tests, it's hard to check anything more than the construction succeeded.
        self.assertEqual(
            SparseObservable.from_raw_parts(
                10, [1.0j, 0.5, 2.0], [1, 3, 2], [0, 1, 2], [0, 1, 1, 3]
            ).num_terms,
            # The three are [(1.0j)*(Z_1), 0.5, 2.0*(X_2 Y_1)]
            3,
        )

    def test_from_raw_parts_checks_coherence(self):
        with self.assertRaisesRegex(ValueError, "not a valid letter"):
            SparseObservable.from_raw_parts(2, [1.0j], [ord("$")], [0], [0, 1])
        with self.assertRaisesRegex(ValueError, r"boundaries.*must be one element longer"):
            SparseObservable.from_raw_parts(2, [1.0j], [SparseObservable.BitTerm.Z], [0], [0])
        with self.assertRaisesRegex(ValueError, r"`bit_terms` \(1\) and `indices` \(0\)"):
            SparseObservable.from_raw_parts(2, [1.0j], [SparseObservable.BitTerm.Z], [], [0, 1])
        with self.assertRaisesRegex(ValueError, r"`bit_terms` \(0\) and `indices` \(1\)"):
            SparseObservable.from_raw_parts(2, [1.0j], [], [1], [0, 1])
        with self.assertRaisesRegex(ValueError, r"the first item of `boundaries` \(1\) must be 0"):
            SparseObservable.from_raw_parts(2, [1.0j], [SparseObservable.BitTerm.Z], [0], [1, 1])
        with self.assertRaisesRegex(ValueError, r"the last item of `boundaries` \(2\)"):
            SparseObservable.from_raw_parts(2, [1.0j], [1], [0], [0, 2])
        with self.assertRaisesRegex(ValueError, r"the last item of `boundaries` \(1\)"):
            SparseObservable.from_raw_parts(2, [1.0j], [1, 2], [0, 1], [0, 1])
        with self.assertRaisesRegex(ValueError, r"all qubit indices must be less than the number"):
            SparseObservable.from_raw_parts(4, [1.0j], [1, 2], [0, 4], [0, 2])
        with self.assertRaisesRegex(ValueError, r"all qubit indices must be less than the number"):
            SparseObservable.from_raw_parts(4, [1.0j, -0.5j], [1, 2], [0, 4], [0, 1, 2])
        with self.assertRaisesRegex(ValueError, "the values in `boundaries` include backwards"):
            SparseObservable.from_raw_parts(
                5, [1.0j, -0.5j, 2.0], [1, 2, 3, 2], [0, 1, 2, 3], [0, 2, 1, 4]
            )
        with self.assertRaisesRegex(
            ValueError, "the values in `indices` are not term-wise increasing"
        ):
            SparseObservable.from_raw_parts(4, [1.0j], [1, 2], [1, 0], [0, 2])

        # There's no test of attempting to pass incoherent data and `check=False` because that
        # permits undefined behaviour in Rust (it's unsafe), so all bets would be off.

    def test_from_label(self):
        # The label is interpreted like a bitstring, with the right-most item associated with qubit
        # 0, and increasing as we move to the left (like `Pauli`, and other bitstring conventions).
        self.assertEqual(
            # Ruler for counting terms:  dcba9876543210
            SparseObservable.from_label("I+-II01IrlIXYZ"),
            SparseObservable.from_raw_parts(
                14,
                [1.0],
                [
                    SparseObservable.BitTerm.Z,
                    SparseObservable.BitTerm.Y,
                    SparseObservable.BitTerm.X,
                    SparseObservable.BitTerm.LEFT,
                    SparseObservable.BitTerm.RIGHT,
                    SparseObservable.BitTerm.ONE,
                    SparseObservable.BitTerm.ZERO,
                    SparseObservable.BitTerm.MINUS,
                    SparseObservable.BitTerm.PLUS,
                ],
                [0, 1, 2, 4, 5, 7, 8, 11, 12],
                [0, 9],
            ),
        )

        self.assertEqual(SparseObservable.from_label("I" * 10), SparseObservable.identity(10))

        # The empty label case is a 0-qubit identity, since `from_label` always sets a coefficient
        # of 1.
        self.assertEqual(SparseObservable.from_label(""), SparseObservable.identity(0))

    def test_from_label_failures(self):
        with self.assertRaisesRegex(ValueError, "labels must only contain letters from"):
            # Bad letters that are still ASCII.
            SparseObservable.from_label("I+-$%I")
        with self.assertRaisesRegex(ValueError, "labels must only contain letters from"):
            # Unicode shenangigans.
            SparseObservable.from_label("🐍")

    def test_from_list(self):
        label = "IXYI+-0lr1ZZY"
        self.assertEqual(
            SparseObservable.from_list([(label, 1.0)]), SparseObservable.from_label(label)
        )
        self.assertEqual(
            SparseObservable.from_list([(label, 1.0)], num_qubits=len(label)),
            SparseObservable.from_label(label),
        )

        self.assertEqual(
            SparseObservable.from_list([("IIIXZI", 1.0j), ("+-IIII", -0.5)]),
            SparseObservable.from_raw_parts(
                6,
                [1.0j, -0.5],
                [
                    SparseObservable.BitTerm.Z,
                    SparseObservable.BitTerm.X,
                    SparseObservable.BitTerm.MINUS,
                    SparseObservable.BitTerm.PLUS,
                ],
                [1, 2, 4, 5],
                [0, 2, 4],
            ),
        )

        self.assertEqual(SparseObservable.from_list([], num_qubits=5), SparseObservable.zero(5))
        self.assertEqual(SparseObservable.from_list([], num_qubits=0), SparseObservable.zero(0))

    def test_from_list_failures(self):
        with self.assertRaisesRegex(ValueError, "labels must only contain letters from"):
            # Bad letters that are still ASCII.
            SparseObservable.from_list([("XZIIZY", 0.5), ("I+-$%I", 1.0j)])
        with self.assertRaisesRegex(ValueError, "labels must only contain letters from"):
            # Unicode shenangigans.
            SparseObservable.from_list([("🐍", 0.5)])
        with self.assertRaisesRegex(ValueError, "label with length 4 cannot be added"):
            SparseObservable.from_list([("IIZ", 0.5), ("IIXI", 1.0j)])
        with self.assertRaisesRegex(ValueError, "label with length 2 cannot be added"):
            SparseObservable.from_list([("IIZ", 0.5), ("II", 1.0j)])
        with self.assertRaisesRegex(ValueError, "label with length 3 cannot be added"):
            SparseObservable.from_list([("IIZ", 0.5), ("IXI", 1.0j)], num_qubits=2)
        with self.assertRaisesRegex(ValueError, "label with length 3 cannot be added"):
            SparseObservable.from_list([("IIZ", 0.5), ("IXI", 1.0j)], num_qubits=4)
        with self.assertRaisesRegex(ValueError, "cannot construct.*without knowing `num_qubits`"):
            SparseObservable.from_list([])

    def test_from_sparse_list(self):
        self.assertEqual(
            SparseObservable.from_sparse_list(
                [
                    ("XY", (0, 1), 0.5),
                    ("+-", (1, 3), -0.25j),
                    ("rl0", (0, 2, 4), 1.0j),
                ],
                num_qubits=5,
            ),
            SparseObservable.from_list([("IIIYX", 0.5), ("I-I+I", -0.25j), ("0IlIr", 1.0j)]),
        )

        # The indices should be allowed to be given in unsorted order, but they should be term-wise
        # sorted in the output.
        from_unsorted = SparseObservable.from_sparse_list(
            [
                ("XYZ", (2, 1, 0), 1.5j),
                ("+rl", (2, 0, 1), -0.5),
            ],
            num_qubits=3,
        )
        self.assertEqual(from_unsorted, SparseObservable.from_list([("XYZ", 1.5j), ("+lr", -0.5)]))
        np.testing.assert_equal(
            from_unsorted.indices, np.array([0, 1, 2, 0, 1, 2], dtype=np.uint32)
        )

        # Explicit identities should still work, just be skipped over.
        explicit_identity = SparseObservable.from_sparse_list(
            [
                ("ZXI", (0, 1, 2), 1.0j),
                ("XYIII", (0, 1, 2, 3, 8), -0.5j),
            ],
            num_qubits=10,
        )
        self.assertEqual(
            explicit_identity,
            SparseObservable.from_sparse_list(
                [("XZ", (1, 0), 1.0j), ("YX", (1, 0), -0.5j)], num_qubits=10
            ),
        )
        np.testing.assert_equal(explicit_identity.indices, np.array([0, 1, 0, 1], dtype=np.uint32))

        self.assertEqual(
            SparseObservable.from_sparse_list([("", (), 1.0)], num_qubits=5),
            SparseObservable.identity(5),
        )
        self.assertEqual(
            SparseObservable.from_sparse_list([("", (), 1.0)], num_qubits=0),
            SparseObservable.identity(0),
        )

        self.assertEqual(
            SparseObservable.from_sparse_list([], num_qubits=1_000_000),
            SparseObservable.zero(1_000_000),
        )
        self.assertEqual(
            SparseObservable.from_sparse_list([], num_qubits=0),
            SparseObservable.zero(0),
        )

    def test_from_sparse_list_failures(self):
        with self.assertRaisesRegex(ValueError, "labels must only contain letters from"):
            # Bad letters that are still ASCII.
            SparseObservable.from_sparse_list(
                [("XZZY", (5, 3, 1, 0), 0.5), ("+$", (2, 1), 1.0j)], num_qubits=8
            )
        # Unicode shenangigans.  These two should fail with a `ValueError`, but the exact message
        # isn't important.  "\xff" is "ÿ", which is two bytes in UTF-8 (so has a length of 2 in
        # Rust), but has a length of 1 in Python, so try with both a length-1 and length-2 index
        # sequence, and both should still raise `ValueError`.
        with self.assertRaises(ValueError):
            SparseObservable.from_sparse_list([("\xff", (1,), 0.5)], num_qubits=5)
        with self.assertRaises(ValueError):
            SparseObservable.from_sparse_list([("\xff", (1, 2), 0.5)], num_qubits=5)

        with self.assertRaisesRegex(ValueError, "label with length 2 does not match indices"):
            SparseObservable.from_sparse_list([("XZ", (0,), 1.0)], num_qubits=5)
        with self.assertRaisesRegex(ValueError, "label with length 2 does not match indices"):
            SparseObservable.from_sparse_list([("XZ", (0, 1, 2), 1.0)], num_qubits=5)

        with self.assertRaisesRegex(ValueError, "index 3 is out of range for a 3-qubit operator"):
            SparseObservable.from_sparse_list([("XZY", (0, 1, 3), 1.0)], num_qubits=3)
        with self.assertRaisesRegex(ValueError, "index 4 is out of range for a 3-qubit operator"):
            SparseObservable.from_sparse_list([("XZY", (0, 1, 4), 1.0)], num_qubits=3)
        with self.assertRaisesRegex(ValueError, "index 3 is out of range for a 3-qubit operator"):
            # ... even if it's for an explicit identity.
            SparseObservable.from_sparse_list([("+-I", (0, 1, 3), 1.0)], num_qubits=3)

        with self.assertRaisesRegex(ValueError, "index 3 is duplicated"):
            SparseObservable.from_sparse_list([("XZ", (3, 3), 1.0)], num_qubits=5)
        with self.assertRaisesRegex(ValueError, "index 3 is duplicated"):
            SparseObservable.from_sparse_list([("XYZXZ", (3, 0, 1, 2, 3), 1.0)], num_qubits=5)

    def test_from_pauli(self):
        # This function should be infallible provided `Pauli` doesn't change its interface and the
        # user doesn't violate the typing.

        # Simple check that the labels are interpreted in the same order.
        self.assertEqual(
            SparseObservable.from_pauli(Pauli("IIXZI")), SparseObservable.from_label("IIXZI")
        )

        # `Pauli` accepts a phase in its label, which we can't (because of clashes with the other
        # alphabet letters), and we should get that right.
        self.assertEqual(
            SparseObservable.from_pauli(Pauli("iIXZIX")),
            SparseObservable.from_list([("IXZIX", 1.0j)]),
        )
        self.assertEqual(
            SparseObservable.from_pauli(Pauli("-iIXZIX")),
            SparseObservable.from_list([("IXZIX", -1.0j)]),
        )
        self.assertEqual(
            SparseObservable.from_pauli(Pauli("-IXZIX")),
            SparseObservable.from_list([("IXZIX", -1.0)]),
        )

        # `Pauli` has its internal phase convention for how it stores `Y`; we should get this right
        # regardless of how many Ys are in the label, or if there's a phase.
        paulis = {"IXYZ" * n: Pauli("IXYZ" * n) for n in range(1, 5)}
        from_paulis, from_labels = zip(
            *(
                (SparseObservable.from_pauli(pauli), SparseObservable.from_label(label))
                for label, pauli in paulis.items()
            )
        )
        self.assertEqual(from_paulis, from_labels)

        phased_paulis = {"IXYZ" * n: Pauli("j" + "IXYZ" * n) for n in range(1, 5)}
        from_paulis, from_lists = zip(
            *(
                (SparseObservable.from_pauli(pauli), SparseObservable.from_list([(label, 1.0j)]))
                for label, pauli in phased_paulis.items()
            )
        )
        self.assertEqual(from_paulis, from_lists)

        self.assertEqual(SparseObservable.from_pauli(Pauli("III")), SparseObservable.identity(3))
        self.assertEqual(SparseObservable.from_pauli(Pauli("")), SparseObservable.identity(0))

    def test_from_sparse_pauli_op(self):
        self.assertEqual(
            SparseObservable.from_sparse_pauli_op(SparsePauliOp.from_list([("IIIII", 1.0)])),
            SparseObservable.identity(5),
        )

        data = [("ZXZXZ", 0.25), ("IYXZI", 1.0j), ("IYYZX", 0.5), ("YYYXI", -0.5), ("IYYYY", 2j)]
        self.assertEqual(
            SparseObservable.from_sparse_pauli_op(SparsePauliOp.from_list(data)),
            SparseObservable.from_list(data),
        )

        # These two _should_ produce the same structure as `SparseObservable.zero(num_qubits)`, but
        # because `SparsePauliOp` doesn't represent the zero operator "natively" - with an empty sum
        # - they actually come out looking like `0.0` times the identity, which is less efficient
        # but acceptable.
        self.assertEqual(
            SparseObservable.from_sparse_pauli_op(SparsePauliOp.from_list([], num_qubits=1)),
            SparseObservable.from_list([("I", 0.0)]),
        )
        self.assertEqual(
            SparseObservable.from_sparse_pauli_op(SparsePauliOp.from_list([], num_qubits=0)),
            SparseObservable.from_list([("", 0.0)]),
        )

    def test_from_sparse_pauli_op_failures(self):
        parametric = SparsePauliOp.from_list([("IIXZ", Parameter("x"))], dtype=object)
        with self.assertRaisesRegex(TypeError, "complex-typed coefficients"):
            SparseObservable.from_sparse_pauli_op(parametric)

    def test_from_terms(self):
        self.assertEqual(SparseObservable.from_terms([], num_qubits=5), SparseObservable.zero(5))
        self.assertEqual(SparseObservable.from_terms((), num_qubits=0), SparseObservable.zero(0))
        self.assertEqual(
            SparseObservable.from_terms((None for _ in []), num_qubits=3), SparseObservable.zero(3)
        )

        expected = SparseObservable.from_sparse_list(
            [
                ("XYZ", (4, 2, 1), 1j),
                ("+-rl", (8, 5, 3, 2), 0.5),
                ("01", (5, 0), 2.0),
            ],
            num_qubits=10,
        )
        self.assertEqual(SparseObservable.from_terms(list(expected)), expected)
        self.assertEqual(SparseObservable.from_terms(tuple(expected)), expected)
        self.assertEqual(SparseObservable.from_terms(term for term in expected), expected)
        self.assertEqual(
            SparseObservable.from_terms(
                (term for term in expected), num_qubits=expected.num_qubits
            ),
            expected,
        )

    def test_from_terms_failures(self):
        with self.assertRaisesRegex(ValueError, "cannot construct.*without knowing `num_qubits`"):
            SparseObservable.from_terms([])

        left, right = SparseObservable("IIXYI")[0], SparseObservable("IIIIIIIIX")[0]
        with self.assertRaisesRegex(ValueError, "mismatched numbers of qubits"):
            SparseObservable.from_terms([left, right])
        with self.assertRaisesRegex(ValueError, "mismatched numbers of qubits"):
            SparseObservable.from_terms([left], num_qubits=100)

    def test_zero(self):
        zero_5 = SparseObservable.zero(5)
        self.assertEqual(zero_5.num_qubits, 5)
        np.testing.assert_equal(zero_5.coeffs, np.array([], dtype=complex))
        np.testing.assert_equal(zero_5.bit_terms, np.array([], dtype=np.uint8))
        np.testing.assert_equal(zero_5.indices, np.array([], dtype=np.uint32))
        np.testing.assert_equal(zero_5.boundaries, np.array([0], dtype=np.uintp))

        zero_0 = SparseObservable.zero(0)
        self.assertEqual(zero_0.num_qubits, 0)
        np.testing.assert_equal(zero_0.coeffs, np.array([], dtype=complex))
        np.testing.assert_equal(zero_0.bit_terms, np.array([], dtype=np.uint8))
        np.testing.assert_equal(zero_0.indices, np.array([], dtype=np.uint32))
        np.testing.assert_equal(zero_0.boundaries, np.array([0], dtype=np.uintp))

    def test_identity(self):
        id_5 = SparseObservable.identity(5)
        self.assertEqual(id_5.num_qubits, 5)
        np.testing.assert_equal(id_5.coeffs, np.array([1], dtype=complex))
        np.testing.assert_equal(id_5.bit_terms, np.array([], dtype=np.uint8))
        np.testing.assert_equal(id_5.indices, np.array([], dtype=np.uint32))
        np.testing.assert_equal(id_5.boundaries, np.array([0, 0], dtype=np.uintp))

        id_0 = SparseObservable.identity(0)
        self.assertEqual(id_0.num_qubits, 0)
        np.testing.assert_equal(id_0.coeffs, np.array([1], dtype=complex))
        np.testing.assert_equal(id_0.bit_terms, np.array([], dtype=np.uint8))
        np.testing.assert_equal(id_0.indices, np.array([], dtype=np.uint32))
        np.testing.assert_equal(id_0.boundaries, np.array([0, 0], dtype=np.uintp))

    @ddt.idata(single_cases())
    def test_copy(self, obs):
        self.assertEqual(obs, obs.copy())
        self.assertIsNot(obs, obs.copy())

    def test_equality(self):
        sparse_data = [("XZ", (1, 0), 0.5j), ("+lr", (3, 1, 0), -0.25)]
        op = SparseObservable.from_sparse_list(sparse_data, num_qubits=5)
        self.assertEqual(op, op.copy())
        # Take care that Rust space allows multiple views onto the same object.
        self.assertEqual(op, op)

        # Comparison to some other object shouldn't fail.
        self.assertNotEqual(op, None)

        # No costly automatic simplification (mathematically, these operators _are_ the same).
        self.assertNotEqual(
            SparseObservable.from_list([("+", 1.0), ("-", 1.0)]), SparseObservable.from_label("X")
        )

        # Difference in qubit count.
        self.assertNotEqual(
            op, SparseObservable.from_sparse_list(sparse_data, num_qubits=op.num_qubits + 1)
        )
        self.assertNotEqual(SparseObservable.zero(2), SparseObservable.zero(3))
        self.assertNotEqual(SparseObservable.identity(2), SparseObservable.identity(3))

        # Difference in coeffs.
        self.assertNotEqual(
            SparseObservable.from_list([("IIXZI", 1.0), ("+-rl0", -0.5j)]),
            SparseObservable.from_list([("IIXZI", 1.0), ("+-rl0", 0.5j)]),
        )
        self.assertNotEqual(
            SparseObservable.from_list([("IIXZI", 1.0), ("+-rl0", -0.5j)]),
            SparseObservable.from_list([("IIXZI", 1.0j), ("+-rl0", -0.5j)]),
        )

        # Difference in bit terms.
        self.assertNotEqual(
            SparseObservable.from_list([("IIXZI", 1.0), ("+-rl0", -0.5j)]),
            SparseObservable.from_list([("IIYZI", 1.0), ("+-rl0", -0.5j)]),
        )
        self.assertNotEqual(
            SparseObservable.from_list([("IIXZI", 1.0), ("+-rl0", -0.5j)]),
            SparseObservable.from_list([("IIXZI", 1.0), ("+-rl1", -0.5j)]),
        )

        # Difference in indices.
        self.assertNotEqual(
            SparseObservable.from_list([("IIXZI", 1.0), ("+Irl0", -0.5j)]),
            SparseObservable.from_list([("IXIZI", 1.0), ("+Irl0", -0.5j)]),
        )
        self.assertNotEqual(
            SparseObservable.from_list([("IIXZI", 1.0), ("+Irl0", -0.5j)]),
            SparseObservable.from_list([("IIXZI", 1.0), ("I+rl0", -0.5j)]),
        )

        # Difference in boundaries.
        self.assertNotEqual(
            SparseObservable.from_sparse_list(
                [("XZ", (0, 1), 1.5), ("+-", (2, 3), -0.5j)], num_qubits=5
            ),
            SparseObservable.from_sparse_list(
                [("XZ+", (0, 1, 2), 1.5), ("-", (3,), -0.5j)], num_qubits=5
            ),
        )

    def test_write_into_attributes_scalar(self):
        coeffs = SparseObservable.from_sparse_list(
            [("XZ", (1, 0), 1.5j), ("+-", (3, 2), -1.5j)], num_qubits=8
        )
        coeffs.coeffs[0] = -2.0
        self.assertEqual(
            coeffs,
            SparseObservable.from_sparse_list(
                [("XZ", (1, 0), -2.0), ("+-", (3, 2), -1.5j)], num_qubits=8
            ),
        )
        coeffs.coeffs[1] = 1.5 + 0.25j
        self.assertEqual(
            coeffs,
            SparseObservable.from_sparse_list(
                [("XZ", (1, 0), -2.0), ("+-", (3, 2), 1.5 + 0.25j)], num_qubits=8
            ),
        )

        bit_terms = SparseObservable.from_sparse_list(
            [("XZ", (0, 1), 1.5j), ("+-", (2, 3), -1.5j)], num_qubits=8
        )
        bit_terms.bit_terms[0] = SparseObservable.BitTerm.Y
        bit_terms.bit_terms[3] = SparseObservable.BitTerm.LEFT
        self.assertEqual(
            bit_terms,
            SparseObservable.from_sparse_list(
                [("YZ", (0, 1), 1.5j), ("+l", (2, 3), -1.5j)], num_qubits=8
            ),
        )

        indices = SparseObservable.from_sparse_list(
            [("XZ", (0, 1), 1.5j), ("+-", (2, 3), -1.5j)], num_qubits=8
        )
        # These two sets keep the observable in term-wise increasing order.  We don't test what
        # happens if somebody violates the Rust-space requirement to be term-wise increasing.
        indices.indices[1] = 4
        indices.indices[3] = 7
        self.assertEqual(
            indices,
            SparseObservable.from_sparse_list(
                [("XZ", (0, 4), 1.5j), ("+-", (2, 7), -1.5j)], num_qubits=8
            ),
        )

        boundaries = SparseObservable.from_sparse_list(
            [("XZ", (0, 1), 1.5j), ("+-", (2, 3), -1.5j)], num_qubits=8
        )
        # Move a single-qubit term from the second summand into the first (the particular indices
        # ensure we remain term-wise sorted).
        boundaries.boundaries[1] += 1
        self.assertEqual(
            boundaries,
            SparseObservable.from_sparse_list(
                [("XZ+", (0, 1, 2), 1.5j), ("-", (3,), -1.5j)], num_qubits=8
            ),
        )

    def test_write_into_attributes_broadcast(self):
        coeffs = SparseObservable.from_list([("XIIZI", 1.5j), ("IIIl0", -0.25), ("1IIIY", 0.5)])
        coeffs.coeffs[:] = 1.5j
        np.testing.assert_array_equal(coeffs.coeffs, [1.5j, 1.5j, 1.5j])
        coeffs.coeffs[1:] = 1.0j
        np.testing.assert_array_equal(coeffs.coeffs, [1.5j, 1.0j, 1.0j])
        coeffs.coeffs[:2] = -0.5
        np.testing.assert_array_equal(coeffs.coeffs, [-0.5, -0.5, 1.0j])
        coeffs.coeffs[::2] = 1.5j
        np.testing.assert_array_equal(coeffs.coeffs, [1.5j, -0.5, 1.5j])
        coeffs.coeffs[::-1] = -0.5j
        np.testing.assert_array_equal(coeffs.coeffs, [-0.5j, -0.5j, -0.5j])

        # It's hard to broadcast into `indices` without breaking data coherence; the broadcasting is
        # more meant for fast modifications to `coeffs` and `bit_terms`.
        indices = SparseObservable.from_list([("XIIZI", 1.5j), ("IIlI0", -0.25), ("1IIIY", 0.5)])
        indices.indices[::2] = 1
        self.assertEqual(
            indices, SparseObservable.from_list([("XIIZI", 1.5j), ("IIl0I", -0.25), ("1IIYI", 0.5)])
        )

        bit_terms = SparseObservable.from_list([("XIIZI", 1.5j), ("IIlI0", -0.25), ("1IIIY", 0.5)])
        bit_terms.bit_terms[::2] = SparseObservable.BitTerm.Z
        self.assertEqual(
            bit_terms,
            SparseObservable.from_list([("XIIZI", 1.5j), ("IIlIZ", -0.25), ("1IIIZ", 0.5)]),
        )
        bit_terms.bit_terms[3:1:-1] = SparseObservable.BitTerm.PLUS
        self.assertEqual(
            bit_terms,
            SparseObservable.from_list([("XIIZI", 1.5j), ("II+I+", -0.25), ("1IIIZ", 0.5)]),
        )
        bit_terms.bit_terms[bit_terms.boundaries[2] : bit_terms.boundaries[3]] = (
            SparseObservable.BitTerm.MINUS
        )
        self.assertEqual(
            bit_terms,
            SparseObservable.from_list([("XIIZI", 1.5j), ("II+I+", -0.25), ("-III-", 0.5)]),
        )

        boundaries = SparseObservable.from_list([("IIIIZX", 1j), ("II+-II", -0.5), ("rlIIII", 0.5)])
        boundaries.boundaries[1:3] = 1
        self.assertEqual(
            boundaries,
            SparseObservable.from_list([("IIIIIX", 1j), ("IIIIII", -0.5), ("rl+-ZI", 0.5)]),
        )

    def test_write_into_attributes_slice(self):
        coeffs = SparseObservable.from_list([("XIIZI", 1.5j), ("IIIl0", -0.25), ("1IIIY", 0.5)])
        coeffs.coeffs[:] = [2.0, 0.5, -0.25]
        self.assertEqual(
            coeffs, SparseObservable.from_list([("XIIZI", 2.0), ("IIIl0", 0.5), ("1IIIY", -0.25)])
        )
        # This should assign the coefficients in reverse order - we more usually spell it
        # `coeffs[:] = coeffs{::-1]`, but the idea is to check the set-item slicing order.
        coeffs.coeffs[::-1] = coeffs.coeffs[:]
        self.assertEqual(
            coeffs, SparseObservable.from_list([("XIIZI", -0.25), ("IIIl0", 0.5), ("1IIIY", 2.0)])
        )

        indices = SparseObservable.from_list([("IIIIZX", 0.25), ("II+-II", 1j), ("rlIIII", 0.5)])
        indices.indices[:4] = [4, 5, 1, 2]
        self.assertEqual(
            indices, SparseObservable.from_list([("ZXIIII", 0.25), ("III+-I", 1j), ("rlIIII", 0.5)])
        )

        bit_terms = SparseObservable.from_list([("IIIIZX", 0.25), ("II+-II", 1j), ("rlIIII", 0.5)])
        bit_terms.bit_terms[::2] = [
            SparseObservable.BitTerm.Y,
            SparseObservable.BitTerm.RIGHT,
            SparseObservable.BitTerm.ZERO,
        ]
        self.assertEqual(
            bit_terms,
            SparseObservable.from_list([("IIIIZY", 0.25), ("II+rII", 1j), ("r0IIII", 0.5)]),
        )

        operators = SparseObservable.from_list([("XZY", 1.5j), ("+1r", -0.5)])
        # Reduce all single-qubit terms to the relevant Pauli operator, if they are a projector.
        operators.bit_terms[:] = operators.bit_terms[:] & 0b00_11
        self.assertEqual(operators, SparseObservable.from_list([("XZY", 1.5j), ("XZY", -0.5)]))

        boundaries = SparseObservable.from_list([("IIIIZX", 0.25), ("II+-II", 1j), ("rlIIII", 0.5)])
        boundaries.boundaries[1:-1] = [1, 5]
        self.assertEqual(
            boundaries,
            SparseObservable.from_list([("IIIIIX", 0.25), ("Il+-ZI", 1j), ("rIIIII", 0.5)]),
        )

    def test_attributes_reject_bad_writes(self):
        obs = SparseObservable.from_list([("XZY", 1.5j), ("+-r", -0.5)])
        with self.assertRaises(TypeError):
            obs.coeffs[0] = [0.25j, 0.5j]
        with self.assertRaises(TypeError):
            obs.bit_terms[0] = [SparseObservable.BitTerm.PLUS] * 4
        with self.assertRaises(TypeError):
            obs.indices[0] = [0, 1]
        with self.assertRaises(TypeError):
            obs.boundaries[0] = (0, 1)
        with self.assertRaisesRegex(ValueError, "not a valid letter"):
            obs.bit_terms[0] = 0
        with self.assertRaisesRegex(ValueError, "not a valid letter"):
            obs.bit_terms[:] = 0
        with self.assertRaisesRegex(
            ValueError, "tried to set a slice of length 2 with a sequence of length 1"
        ):
            obs.coeffs[:] = [1.0j]
        with self.assertRaisesRegex(
            ValueError, "tried to set a slice of length 6 with a sequence of length 8"
        ):
            obs.bit_terms[:] = [SparseObservable.BitTerm.Z] * 8

    def test_attributes_sequence(self):
        """Test attributes of the `Sequence` protocol."""
        # Length
        obs = SparseObservable.from_list([("XZY", 1.5j), ("+-r", -0.5)])
        self.assertEqual(len(obs.coeffs), 2)
        self.assertEqual(len(obs.indices), 6)
        self.assertEqual(len(obs.bit_terms), 6)
        self.assertEqual(len(obs.boundaries), 3)

        # Iteration
        self.assertEqual(list(obs.coeffs), [1.5j, -0.5])
        self.assertEqual(tuple(obs.indices), (0, 1, 2, 0, 1, 2))
        self.assertEqual(next(iter(obs.boundaries)), 0)
        # multiple iteration through same object
        bit_terms = obs.bit_terms
        self.assertEqual(set(bit_terms), {SparseObservable.BitTerm[x] for x in "XYZ+-r"})
        self.assertEqual(set(bit_terms), {SparseObservable.BitTerm[x] for x in "XYZ+-r"})

        # Implicit iteration methods.
        self.assertIn(SparseObservable.BitTerm.PLUS, obs.bit_terms)
        self.assertNotIn(4, obs.indices)
        self.assertEqual(list(reversed(obs.coeffs)), [-0.5, 1.5j])

        # Index by scalar
        self.assertEqual(obs.coeffs[1], -0.5)
        self.assertEqual(obs.indices[-1], 2)
        self.assertEqual(obs.bit_terms[0], SparseObservable.BitTerm.Y)
        # Make sure that Rust-space actually returns the enum value, not just an `int` (which could
        # have compared equal).
        self.assertIsInstance(obs.bit_terms[0], SparseObservable.BitTerm)
        self.assertEqual(obs.boundaries[-2], 3)
        with self.assertRaises(IndexError):
            _ = obs.coeffs[10]
        with self.assertRaises(IndexError):
            _ = obs.boundaries[-4]

        # Index by slice.  This is API guaranteed to be a Numpy array to make it easier to
        # manipulate subslices with mathematic operations.
        self.assertIsInstance(obs.coeffs[:], np.ndarray)
        np.testing.assert_array_equal(
            obs.coeffs[:], np.array([1.5j, -0.5], dtype=np.complex128), strict=True
        )
        self.assertIsInstance(obs.indices[::-1], np.ndarray)
        np.testing.assert_array_equal(
            obs.indices[::-1], np.array([2, 1, 0, 2, 1, 0], dtype=np.uint32), strict=True
        )
        self.assertIsInstance(obs.bit_terms[2:4], np.ndarray)
        np.testing.assert_array_equal(
            obs.bit_terms[2:4],
            np.array([SparseObservable.BitTerm.X, SparseObservable.BitTerm.RIGHT], dtype=np.uint8),
            strict=True,
        )
        self.assertIsInstance(obs.boundaries[-2:-3:-1], np.ndarray)
        np.testing.assert_array_equal(
            obs.boundaries[-2:-3:-1], np.array([3], dtype=np.uintp), strict=True
        )

    def test_attributes_to_array(self):
        obs = SparseObservable.from_list([("XZY", 1.5j), ("+-r", -0.5)])

        # Natural dtypes.
        np.testing.assert_array_equal(
            obs.coeffs, np.array([1.5j, -0.5], dtype=np.complex128), strict=True
        )
        np.testing.assert_array_equal(
            obs.indices, np.array([0, 1, 2, 0, 1, 2], dtype=np.uint32), strict=True
        )
        np.testing.assert_array_equal(
            obs.bit_terms,
            np.array([SparseObservable.BitTerm[x] for x in "YZXr-+"], dtype=np.uint8),
            strict=True,
        )
        np.testing.assert_array_equal(
            obs.boundaries, np.array([0, 3, 6], dtype=np.uintp), strict=True
        )

        # Cast dtypes.
        np.testing.assert_array_equal(
            np.array(obs.indices, dtype=np.uint8),
            np.array([0, 1, 2, 0, 1, 2], dtype=np.uint8),
            strict=True,
        )
        np.testing.assert_array_equal(
            np.array(obs.boundaries, dtype=np.int64),
            np.array([0, 3, 6], dtype=np.int64),
            strict=True,
        )

    @unittest.skipIf(
        int(np.__version__.split(".", maxsplit=1)[0]) < 2,
        "Numpy 1.x did not have a 'copy' keyword parameter to 'numpy.asarray'",
    )
    def test_attributes_reject_no_copy_array(self):
        obs = SparseObservable.from_list([("XZY", 1.5j), ("+-r", -0.5)])
        with self.assertRaisesRegex(ValueError, "cannot produce a safe view"):
            np.asarray(obs.coeffs, copy=False)
        with self.assertRaisesRegex(ValueError, "cannot produce a safe view"):
            np.asarray(obs.indices, copy=False)
        with self.assertRaisesRegex(ValueError, "cannot produce a safe view"):
            np.asarray(obs.bit_terms, copy=False)
        with self.assertRaisesRegex(ValueError, "cannot produce a safe view"):
            np.asarray(obs.boundaries, copy=False)

    def test_attributes_repr(self):
        # We're not testing much about the outputs here, just that they don't crash.
        obs = SparseObservable.from_list([("XZY", 1.5j), ("+-r", -0.5)])
        self.assertIn("coeffs", repr(obs.coeffs))
        self.assertIn("bit_terms", repr(obs.bit_terms))
        self.assertIn("indices", repr(obs.indices))
        self.assertIn("boundaries", repr(obs.boundaries))

    @combine(
        obs=single_cases(),
        # This includes some elements that aren't native `complex`, but still should be cast.
        coeff=[0.5, 3j, 2, 0.25 - 0.75j],
    )
    def test_multiply(self, obs, coeff):
        obs = obs.copy()
        initial = obs.copy()
        expected = obs.copy()
        expected.coeffs[:] = np.asarray(expected.coeffs) * complex(coeff)
        self.assertEqual(obs * coeff, expected)
        self.assertEqual(coeff * obs, expected)
        # Check that nothing applied in-place.
        self.assertEqual(obs, initial)
        obs *= coeff
        self.assertEqual(obs, expected)
        self.assertIs(obs * AllowRightArithmetic(), AllowRightArithmetic.SENTINEL)

    @ddt.idata(single_cases())
    def test_multiply_zero(self, obs):
        initial = obs.copy()
        self.assertEqual(obs * 0.0, SparseObservable.zero(initial.num_qubits))
        self.assertEqual(0.0 * obs, SparseObservable.zero(initial.num_qubits))
        self.assertEqual(obs, initial)

        obs *= 0.0
        self.assertEqual(obs, SparseObservable.zero(initial.num_qubits))

    @combine(
        obs=single_cases(),
        # This includes some elements that aren't native `complex`, but still should be cast.  Be
        # careful that the floating-point operation should not involve rounding.
        coeff=[0.5, 4j, 2, -0.25],
    )
    def test_divide(self, obs, coeff):
        obs = obs.copy()
        initial = obs.copy()
        expected = obs.copy()
        expected.coeffs[:] = np.asarray(expected.coeffs) / complex(coeff)
        self.assertEqual(obs / coeff, expected)
        # Check that nothing applied in-place.
        self.assertEqual(obs, initial)
        obs /= coeff
        self.assertEqual(obs, expected)
        self.assertIs(obs / AllowRightArithmetic(), AllowRightArithmetic.SENTINEL)

    @ddt.idata(single_cases())
    def test_divide_zero_raises(self, obs):
        with self.assertRaises(ZeroDivisionError):
            _ = obs / 0.0j
        with self.assertRaises(ZeroDivisionError):
            obs /= 0.0j

    def test_add_simple(self):
        num_qubits = 12
        terms = [
            ("ZXY", (5, 2, 1), 1.5j),
            ("+r", (8, 0), -0.25),
            ("-0l1", (10, 9, 4, 3), 0.5 + 1j),
            ("XZ", (7, 5), 0.75j),
            ("rl01", (5, 3, 1, 0), 0.25j),
        ]
        expected = SparseObservable.from_sparse_list(terms, num_qubits=num_qubits)
        for pivot in range(1, len(terms) - 1):
            left = SparseObservable.from_sparse_list(terms[:pivot], num_qubits=num_qubits)
            left_initial = left.copy()
            right = SparseObservable.from_sparse_list(terms[pivot:], num_qubits=num_qubits)
            right_initial = right.copy()
            # Addition is documented to be term-stacking, so structural equality without `simplify`
            # should hold.
            self.assertEqual(left + right, expected)
            # This is a different order, so check the simplification and canonicalisation works.
            self.assertEqual((right + left).simplify(), expected.simplify())
            # Neither was modified in place.
            self.assertEqual(left, left_initial)
            self.assertEqual(right, right_initial)

            left += right
            self.assertEqual(left, expected)
            self.assertEqual(right, right_initial)

    @ddt.idata(single_cases())
    def test_add_self(self, obs):
        """Test that addition to `self` works fine, including in-place mutation.  This is a case
        where we might fall afoul of Rust's borrowing rules."""
        initial = obs.copy()
        expected = (2.0 * obs).simplify()
        self.assertEqual((obs + obs).simplify(), expected)
        self.assertEqual(obs, initial)

        obs += obs
        self.assertEqual(obs.simplify(), expected)

    @ddt.idata(single_cases())
    def test_add_zero(self, obs):
        expected = obs.copy()
        zero = SparseObservable.zero(obs.num_qubits)
        self.assertEqual(obs + zero, expected)
        self.assertEqual(zero + obs, expected)

        obs += zero
        self.assertEqual(obs, expected)
        zero += obs
        self.assertEqual(zero, expected)

    def test_add_coercion(self):
        """Other quantum-info operators coerce with the ``+`` operator, so we do too."""
        base = SparseObservable.zero(9)

        pauli_label = "IIIXYZIII"
        expected = SparseObservable.from_label(pauli_label)
        self.assertEqual(base + pauli_label, expected)
        self.assertEqual(pauli_label + base, expected)

        pauli = Pauli(pauli_label)
        self.assertEqual(base + pauli, expected)
        self.assertEqual(pauli + base, expected)

        spo = SparsePauliOp(pauli_label)
        self.assertEqual(base + spo, expected)
        with self.assertRaisesRegex(QiskitError, "Invalid input data for Pauli"):
            # This doesn't work because `SparsePauliOp` is badly behaved in its coercion (it gets
            # first dibs at `__add__`, not our `__radd__`), and will not return `NotImplemented` for
            # bad types.  This _shouldn't_ raise, and this test here is to remind us to flip it to a
            # proper assertion of correctness if `Pauli` starts playing nicely.
            _ = spo + base

        obs_label = "10+-rlXYZ"
        expected = SparseObservable.from_label(obs_label)
        self.assertEqual(base + obs_label, expected)
        self.assertEqual(obs_label + base, expected)

        expected = 3j * SparseObservable.from_label("IXYrlII0I")
        self.assertEqual(base + expected[0], expected)
        self.assertEqual(expected[0] + base, expected)

        with self.assertRaises(TypeError):
            _ = base + {}
        with self.assertRaises(TypeError):
            _ = {} + base
        with self.assertRaisesRegex(ValueError, "only contain letters from the alphabet"):
            _ = base + "$$$"
        with self.assertRaisesRegex(ValueError, "only contain letters from the alphabet"):
            _ = "$$$" + base

        self.assertIs(base + AllowRightArithmetic(), AllowRightArithmetic.SENTINEL)
        with self.assertRaisesRegex(TypeError, "invalid object for in-place addition"):
            # This actually _shouldn't_ be a `TypeError` - `__iadd_` should defer to
            # `AllowRightArithmetic.__radd__` in the same way that `__add__` does, but a limitation
            # in PyO3 (see PyO3/pyo3#4605) prevents this.
            base += AllowRightArithmetic()

    def test_add_failures(self):
        with self.assertRaisesRegex(ValueError, "incompatible numbers of qubits"):
            _ = SparseObservable.zero(4) + SparseObservable.zero(6)
        with self.assertRaisesRegex(ValueError, "incompatible numbers of qubits"):
            _ = SparseObservable.zero(6) + SparseObservable.zero(4)

    def test_sub_simple(self):
        num_qubits = 12
        terms = [
            ("ZXY", (5, 2, 1), 1.5j),
            ("+r", (8, 0), -0.25),
            ("-0l1", (10, 9, 4, 3), 0.5 + 1j),
            ("XZ", (7, 5), 0.75j),
            ("rl01", (5, 3, 1, 0), 0.25j),
        ]
        for pivot in range(1, len(terms) - 1):
            expected = SparseObservable.from_sparse_list(
                [
                    (label, indices, coeff if i < pivot else -coeff)
                    for i, (label, indices, coeff) in enumerate(terms)
                ],
                num_qubits=num_qubits,
            )
            left = SparseObservable.from_sparse_list(terms[:pivot], num_qubits=num_qubits)
            left_initial = left.copy()
            right = SparseObservable.from_sparse_list(terms[pivot:], num_qubits=num_qubits)
            right_initial = right.copy()
            # Addition is documented to be term-stacking, so structural equality without `simplify`
            # should hold.
            self.assertEqual(left - right, expected)
            # This is a different order, so check the simplification and canonicalisation works.
            self.assertEqual((right - left).simplify(), -expected.simplify())
            # Neither was modified in place.
            self.assertEqual(left, left_initial)
            self.assertEqual(right, right_initial)

            left -= right
            self.assertEqual(left, expected)
            self.assertEqual(right, right_initial)

    @ddt.idata(single_cases())
    def test_sub_self(self, obs):
        """Test that subtraction of `self` works fine, including in-place mutation.  This is a case
        where we might fall afoul of Rust's borrowing rules."""
        initial = obs.copy()
        expected = SparseObservable.zero(obs.num_qubits)
        self.assertEqual((obs - obs).simplify(), expected)
        self.assertEqual(obs, initial)

        obs -= obs
        self.assertEqual(obs.simplify(), expected)

    @ddt.idata(single_cases())
    def test_sub_zero(self, obs):
        expected = obs.copy()
        zero = SparseObservable.zero(obs.num_qubits)
        self.assertEqual(obs - zero, expected)
        self.assertEqual(zero - obs, -expected)

        obs -= zero
        self.assertEqual(obs, expected)
        zero -= obs
        self.assertEqual(zero, -expected)

    def test_sub_coercion(self):
        """Other quantum-info operators coerce with the ``-`` operator, so we do too."""
        base = SparseObservable.zero(9)

        pauli_label = "IIIXYZIII"
        expected = SparseObservable.from_label(pauli_label)
        self.assertEqual(base - pauli_label, -expected)
        self.assertEqual(pauli_label - base, expected)

        pauli = Pauli(pauli_label)
        self.assertEqual(base - pauli, -expected)
        self.assertEqual(pauli - base, expected)

        spo = SparsePauliOp(pauli_label)
        self.assertEqual(base - spo, -expected)
        with self.assertRaisesRegex(QiskitError, "Invalid input data for Pauli"):
            # This doesn't work because `SparsePauliOp` is badly behaved in its coercion (it gets
            # first dibs at `__add__`, not our `__radd__`), and will not return `NotImplemented` for
            # bad types.  This _shouldn't_ raise, and this test here is to remind us to flip it to a
            # proper assertion of correctness if `Pauli` starts playing nicely.
            _ = spo + base

        obs_label = "10+-rlXYZ"
        expected = SparseObservable.from_label(obs_label)
        self.assertEqual(base - obs_label, -expected)
        self.assertEqual(obs_label - base, expected)

        expected = 3j * SparseObservable.from_label("IXYrlII0I")
        self.assertEqual(base - expected[0], -expected)
        self.assertEqual(expected[0] - base, expected)

        with self.assertRaises(TypeError):
            _ = base - {}
        with self.assertRaises(TypeError):
            _ = {} - base
        with self.assertRaisesRegex(ValueError, "only contain letters from the alphabet"):
            _ = base - "$$$"
        with self.assertRaisesRegex(ValueError, "only contain letters from the alphabet"):
            _ = "$$$" - base

        self.assertIs(base + AllowRightArithmetic(), AllowRightArithmetic.SENTINEL)
        with self.assertRaisesRegex(TypeError, "invalid object for in-place subtraction"):
            # This actually _shouldn't_ be a `TypeError` - `__isub_` should defer to
            # `AllowRightArithmetic.__rsub__` in the same way that `__sub__` does, but a limitation
            # in PyO3 (see PyO3/pyo3#4605) prevents this.
            base -= AllowRightArithmetic()

    def test_sub_failures(self):
        with self.assertRaisesRegex(ValueError, "incompatible numbers of qubits"):
            _ = SparseObservable.zero(4) - SparseObservable.zero(6)
        with self.assertRaisesRegex(ValueError, "incompatible numbers of qubits"):
            _ = SparseObservable.zero(6) - SparseObservable.zero(4)

    @ddt.idata(single_cases())
    def test_neg(self, obs):
        initial = obs.copy()
        expected = obs.copy()
        expected.coeffs[:] = -np.asarray(expected.coeffs)
        self.assertEqual(-obs, expected)
        # Test that there's no in-place modification.
        self.assertEqual(obs, initial)

    @ddt.idata(single_cases())
    def test_pos(self, obs):
        initial = obs.copy()
        self.assertEqual(+obs, initial)
        self.assertIsNot(+obs, obs)

    @combine(left=single_cases(), right=single_cases())
    def test_tensor(self, left, right):

        def expected(left, right):
            coeffs = []
            bit_terms = []
            indices = []
            boundaries = [0]
            for left_ptr in range(left.num_terms):
                left_start, left_end = left.boundaries[left_ptr], left.boundaries[left_ptr + 1]
                for right_ptr in range(right.num_terms):
                    right_start = right.boundaries[right_ptr]
                    right_end = right.boundaries[right_ptr + 1]
                    coeffs.append(left.coeffs[left_ptr] * right.coeffs[right_ptr])
                    bit_terms.extend(right.bit_terms[right_start:right_end])
                    bit_terms.extend(left.bit_terms[left_start:left_end])
                    indices.extend(right.indices[right_start:right_end])
                    indices.extend(i + right.num_qubits for i in left.indices[left_start:left_end])
                    boundaries.append(len(indices))
            return SparseObservable.from_raw_parts(
                left.num_qubits + right.num_qubits, coeffs, bit_terms, indices, boundaries
            )

        # We deliberately have the arguments flipped when appropriate, here.
        # pylint: disable=arguments-out-of-order

        left_initial = left.copy()
        right_initial = right.copy()
        self.assertEqual(left.tensor(right), expected(left, right))
        self.assertEqual(left, left_initial)
        self.assertEqual(right, right_initial)
        self.assertEqual(right.tensor(left), expected(right, left))

        self.assertEqual(left.expand(right), expected(right, left))
        self.assertEqual(left, left_initial)
        self.assertEqual(right, right_initial)
        self.assertEqual(right.expand(left), expected(left, right))

        self.assertEqual(left.tensor(right), right.expand(left))
        self.assertEqual(left.expand(right), right.tensor(left))

    @combine(
        obs=single_cases(), identity=[SparseObservable.identity(0), SparseObservable.identity(5)]
    )
    def test_tensor_identity(self, obs, identity):
        initial = obs.copy()
        expected_left = SparseObservable.from_raw_parts(
            obs.num_qubits + identity.num_qubits,
            obs.coeffs,
            obs.bit_terms,
            [x + identity.num_qubits for x in obs.indices],
            obs.boundaries,
        )
        expected_right = SparseObservable.from_raw_parts(
            obs.num_qubits + identity.num_qubits,
            obs.coeffs,
            obs.bit_terms,
            obs.indices,
            obs.boundaries,
        )
        self.assertEqual(obs.tensor(identity), expected_left)
        self.assertEqual(identity.tensor(obs), expected_right)
        self.assertEqual(obs.expand(identity), expected_right)
        self.assertEqual(identity.expand(obs), expected_left)
        self.assertEqual(obs ^ identity, expected_left)
        self.assertEqual(identity ^ obs, expected_right)
        self.assertEqual(obs, initial)
        obs ^= identity
        self.assertEqual(obs, expected_left)

    @combine(obs=single_cases(), zero=[SparseObservable.zero(0), SparseObservable.zero(5)])
    def test_tensor_zero(self, obs, zero):
        initial = obs.copy()
        expected = SparseObservable.zero(obs.num_qubits + zero.num_qubits)
        self.assertEqual(obs.tensor(zero), expected)
        self.assertEqual(zero.tensor(obs), expected)
        self.assertEqual(obs.expand(zero), expected)
        self.assertEqual(zero.expand(obs), expected)
        self.assertEqual(obs ^ zero, expected)
        self.assertEqual(zero ^ obs, expected)
        self.assertEqual(obs, initial)
        obs ^= zero
        self.assertEqual(obs, expected)

    def test_tensor_coercion(self):
        """Other quantum-info operators coerce with the ``tensor`` method and operator, so we do
        too."""
        base = SparseObservable.identity(0)

        pauli_label = "IIXYZII"
        expected = SparseObservable.from_label(pauli_label)
        self.assertEqual(base.tensor(pauli_label), expected)
        self.assertEqual(base.expand(pauli_label), expected)
        self.assertEqual(base ^ pauli_label, expected)
        self.assertEqual(pauli_label ^ base, expected)

        pauli = Pauli(pauli_label)
        self.assertEqual(base.tensor(pauli), expected)
        self.assertEqual(base.expand(pauli), expected)
        self.assertEqual(base ^ pauli, expected)
        with self.assertRaisesRegex(QiskitError, "Invalid input data for Pauli"):
            # This doesn't work because `Pauli` is badly behaved in its coercion (it gets first dibs
            # at `__xor__`, not our `__rxor__`), and will not return `NotImplemented` for bad types.
            # This _shouldn't_ raise, and this test here is to remind us to flip it to a proper
            # assertion of correctness if `Pauli` starts playing nicely.
            _ = pauli ^ base

        spo = SparsePauliOp(pauli_label)
        self.assertEqual(base.tensor(spo), expected)
        self.assertEqual(base.expand(spo), expected)
        self.assertEqual(base ^ spo, expected)
        with self.assertRaisesRegex(QiskitError, "Invalid input data for Pauli"):
            # This doesn't work because `SparsePauliOp` is badly behaved in its coercion (it gets
            # first dibs at `__xor__`, not our `__rxor__`), and will not return `NotImplemented` for
            # bad types.  This _shouldn't_ raise, and this test here is to remind us to flip it to a
            # proper assertion of correctness if `Pauli` starts playing nicely.
            _ = spo ^ base

        obs_label = "10+-rlXYZ"
        expected = SparseObservable.from_label(obs_label)
        self.assertEqual(base.tensor(obs_label), expected)
        self.assertEqual(base.expand(obs_label), expected)
        self.assertEqual(base ^ obs_label, expected)
        self.assertEqual(obs_label ^ base, expected)

        with self.assertRaises(TypeError):
            _ = base ^ {}
        with self.assertRaises(TypeError):
            _ = {} ^ base
        with self.assertRaisesRegex(ValueError, "only contain letters from the alphabet"):
            _ = base ^ "$$$"
        with self.assertRaisesRegex(ValueError, "only contain letters from the alphabet"):
            _ = "$$$" ^ base

        self.assertIs(base ^ AllowRightArithmetic(), AllowRightArithmetic.SENTINEL)

    @ddt.idata(single_cases())
    def test_adjoint(self, obs):
        initial = obs.copy()
        expected = obs.copy()
        expected.coeffs[:] = np.conjugate(expected.coeffs)
        self.assertEqual(obs.adjoint(), expected)
        self.assertEqual(obs, initial)
        self.assertEqual(obs.adjoint().adjoint(), initial)
        self.assertEqual(obs.adjoint(), obs.conjugate().transpose())
        self.assertEqual(obs.adjoint(), obs.transpose().conjugate())

    @ddt.idata(single_cases())
    def test_conjugate(self, obs):
        initial = obs.copy()

        term_map = {term: (term, 1.0) for term in SparseObservable.BitTerm}
        term_map[SparseObservable.BitTerm.Y] = (SparseObservable.BitTerm.Y, -1.0)
        term_map[SparseObservable.BitTerm.RIGHT] = (SparseObservable.BitTerm.LEFT, 1.0)
        term_map[SparseObservable.BitTerm.LEFT] = (SparseObservable.BitTerm.RIGHT, 1.0)

        expected = obs.copy()
        for i in range(expected.num_terms):
            start, end = expected.boundaries[i], expected.boundaries[i + 1]
            coeff = expected.coeffs[i]
            for offset, bit_term in enumerate(expected.bit_terms[start:end]):
                new_term, multiplier = term_map[bit_term]
                coeff *= multiplier
                expected.bit_terms[start + offset] = new_term
            expected.coeffs[i] = coeff.conjugate()

        self.assertEqual(obs.conjugate(), expected)
        self.assertEqual(obs, initial)
        self.assertEqual(obs.conjugate().conjugate(), initial)
        self.assertEqual(obs.conjugate(), obs.transpose().adjoint())
        self.assertEqual(obs.conjugate(), obs.adjoint().transpose())

    def test_conjugate_explicit(self):
        # The description of conjugation on the operator is not 100% trivial to see is correct, so
        # here's an explicit case to verify.
        obs = SparseObservable.from_sparse_list(
            [
                ("Y", (1,), 2.0),
                ("X+-", (5, 4, 3), 1.5),
                ("Z01", (5, 4, 3), 1.5j),
                ("YY", (2, 0), 0.25),
                ("YY", (3, 1), 0.25j),
                ("YYY", (3, 2, 1), 0.75),
                ("rlrl", (4, 3, 2, 1), 1.0),
                ("lrlr", (4, 3, 2, 1), 1.0j),
                ("", (), 1.5j),
            ],
            num_qubits=6,
        )
        expected = SparseObservable.from_sparse_list(
            [
                ("Y", (1,), -2.0),
                ("X+-", (5, 4, 3), 1.5),
                ("Z01", (5, 4, 3), -1.5j),
                ("YY", (2, 0), 0.25),
                ("YY", (3, 1), -0.25j),
                ("YYY", (3, 2, 1), -0.75),
                ("lrlr", (4, 3, 2, 1), 1.0),
                ("rlrl", (4, 3, 2, 1), -1.0j),
                ("", (), -1.5j),
            ],
            num_qubits=6,
        )
        self.assertEqual(obs.conjugate(), expected)
        self.assertEqual(obs.conjugate().conjugate(), obs)

    @ddt.idata(single_cases())
    def test_transpose(self, obs):
        initial = obs.copy()

        term_map = {term: (term, 1.0) for term in SparseObservable.BitTerm}
        term_map[SparseObservable.BitTerm.Y] = (SparseObservable.BitTerm.Y, -1.0)
        term_map[SparseObservable.BitTerm.RIGHT] = (SparseObservable.BitTerm.LEFT, 1.0)
        term_map[SparseObservable.BitTerm.LEFT] = (SparseObservable.BitTerm.RIGHT, 1.0)

        expected = obs.copy()
        for i in range(expected.num_terms):
            start, end = expected.boundaries[i], expected.boundaries[i + 1]
            coeff = expected.coeffs[i]
            for offset, bit_term in enumerate(expected.bit_terms[start:end]):
                new_term, multiplier = term_map[bit_term]
                coeff *= multiplier
                expected.bit_terms[start + offset] = new_term
            expected.coeffs[i] = coeff

        self.assertEqual(obs.transpose(), expected)
        self.assertEqual(obs, initial)
        self.assertEqual(obs.transpose().transpose(), initial)
        self.assertEqual(obs.transpose(), obs.conjugate().adjoint())
        self.assertEqual(obs.transpose(), obs.adjoint().conjugate())

    def test_transpose_explicit(self):
        # The description of transposition on the operator is not 100% trivial to see is correct, so
        # here's a few explicit cases to verify.
        obs = SparseObservable.from_sparse_list(
            [
                ("Y", (1,), 2.0),
                ("X+-", (5, 4, 3), 1.5),
                ("Z01", (5, 4, 3), 1.5j),
                ("YY", (2, 0), 0.25),
                ("YY", (3, 1), 0.25j),
                ("YYY", (3, 2, 1), 0.75),
                ("rlrl", (4, 3, 2, 1), 1.0),
                ("lrlr", (4, 3, 2, 1), 1.0j),
                ("", (), 1.5j),
            ],
            num_qubits=6,
        )
        expected = SparseObservable.from_sparse_list(
            [
                ("Y", (1,), -2.0),
                ("X+-", (5, 4, 3), 1.5),
                ("Z01", (5, 4, 3), 1.5j),
                ("YY", (2, 0), 0.25),
                ("YY", (3, 1), 0.25j),
                ("YYY", (3, 2, 1), -0.75),
                ("lrlr", (4, 3, 2, 1), 1.0),
                ("rlrl", (4, 3, 2, 1), 1.0j),
                ("", (), 1.5j),
            ],
            num_qubits=6,
        )
        self.assertEqual(obs.transpose(), expected)
        self.assertEqual(obs.transpose().transpose(), obs)

    def test_simplify(self):
        self.assertEqual((1e-10 * SparseObservable("XX")).simplify(1e-8), SparseObservable.zero(2))
        self.assertEqual((1e-10j * SparseObservable("XX")).simplify(1e-8), SparseObservable.zero(2))
        self.assertEqual(
            (1e-7 * SparseObservable("XX")).simplify(1e-8), 1e-7 * SparseObservable("XX")
        )

        exact_coeff = 2.0**-10
        self.assertEqual(
            (exact_coeff * SparseObservable("XX")).simplify(exact_coeff), SparseObservable.zero(2)
        )
        self.assertEqual(
            (exact_coeff * 1j * SparseObservable("XX")).simplify(exact_coeff),
            SparseObservable.zero(2),
        )
        coeff = 3e-5 + 4e-5j
        self.assertEqual(
            (coeff * SparseObservable("ZZ")).simplify(abs(coeff)), SparseObservable.zero(2)
        )

        sum_alike = SparseObservable.from_list(
            [
                ("XX", 1.0),
                ("YY", 1j),
                ("XX", -1.0),
            ]
        )
        self.assertEqual(sum_alike.simplify(), 1j * SparseObservable("YY"))

        terms = [
            ("XYIZI", 1.5),
            ("+-IYI", 2.0),
            ("XYIZI", 2j),
            ("+-IYI", -2.0),
            ("rlIZI", -2.0),
        ]
        canonical_forwards = SparseObservable.from_list(terms)
        canonical_backwards = SparseObservable.from_list(list(reversed(terms)))
        self.assertNotEqual(canonical_forwards.simplify(), canonical_forwards)
        self.assertNotEqual(canonical_forwards, canonical_backwards)
        self.assertEqual(canonical_forwards.simplify(), canonical_backwards.simplify())
        self.assertEqual(canonical_forwards.simplify(), canonical_forwards.simplify().simplify())

    @ddt.idata(single_cases())
    def test_clear(self, obs):
        num_qubits = obs.num_qubits
        obs.clear()
        self.assertEqual(obs, SparseObservable.zero(num_qubits))

    def test_apply_layout_list(self):
        self.assertEqual(
            SparseObservable.zero(5).apply_layout([4, 3, 2, 1, 0]), SparseObservable.zero(5)
        )
        self.assertEqual(
            SparseObservable.zero(3).apply_layout([0, 2, 1], 8), SparseObservable.zero(8)
        )
        self.assertEqual(
            SparseObservable.identity(2).apply_layout([1, 0]), SparseObservable.identity(2)
        )
        self.assertEqual(
            SparseObservable.identity(3).apply_layout([100, 10_000, 3], 100_000_000),
            SparseObservable.identity(100_000_000),
        )

        terms = [
            ("ZYX", (4, 2, 1), 1j),
            ("", (), -0.5),
            ("+-rl01", (10, 8, 6, 4, 2, 0), 2.0),
        ]

        def map_indices(terms, layout):
            return [
                (terms, tuple(layout[bit] for bit in bits), coeff) for terms, bits, coeff in terms
            ]

        identity = list(range(12))
        self.assertEqual(
            SparseObservable.from_sparse_list(terms, num_qubits=12).apply_layout(identity),
            SparseObservable.from_sparse_list(terms, num_qubits=12),
        )
        # We've already tested elsewhere that `SparseObservable.from_sparse_list` produces termwise
        # sorted indices, so these tests also ensure `apply_layout` is maintaining that invariant.
        backwards = list(range(12))[::-1]
        self.assertEqual(
            SparseObservable.from_sparse_list(terms, num_qubits=12).apply_layout(backwards),
            SparseObservable.from_sparse_list(map_indices(terms, backwards), num_qubits=12),
        )
        shuffled = [4, 7, 1, 10, 0, 11, 3, 2, 8, 5, 6, 9]
        self.assertEqual(
            SparseObservable.from_sparse_list(terms, num_qubits=12).apply_layout(shuffled),
            SparseObservable.from_sparse_list(map_indices(terms, shuffled), num_qubits=12),
        )
        self.assertEqual(
            SparseObservable.from_sparse_list(terms, num_qubits=12).apply_layout(shuffled, 100),
            SparseObservable.from_sparse_list(map_indices(terms, shuffled), num_qubits=100),
        )
        expanded = [78, 69, 82, 68, 32, 97, 108, 101, 114, 116, 33]
        self.assertEqual(
            SparseObservable.from_sparse_list(terms, num_qubits=11).apply_layout(expanded, 120),
            SparseObservable.from_sparse_list(map_indices(terms, expanded), num_qubits=120),
        )

    def test_apply_layout_transpiled(self):
        base = SparseObservable.from_sparse_list(
            [
                ("ZYX", (4, 2, 1), 1j),
                ("", (), -0.5),
                ("+-r", (3, 2, 0), 2.0),
            ],
            num_qubits=5,
        )

        qc = QuantumCircuit(5)
        initial_list = [3, 4, 0, 2, 1]
        no_routing = transpile(
            qc, target=lnn_target(5), initial_layout=initial_list, seed_transpiler=2024_10_25_0
        ).layout
        # It's easiest here to test against the `list` form, which we verify separately and
        # explicitly.
        self.assertEqual(base.apply_layout(no_routing), base.apply_layout(initial_list))

        expanded = transpile(
            qc, target=lnn_target(100), initial_layout=initial_list, seed_transpiler=2024_10_25_1
        ).layout
        self.assertEqual(
            base.apply_layout(expanded), base.apply_layout(initial_list, num_qubits=100)
        )

        qc = QuantumCircuit(5)
        qargs = list(itertools.permutations(range(5), 2))
        random.Random(2024_10_25_2).shuffle(qargs)
        for pair in qargs:
            qc.cx(*pair)

        routed = transpile(qc, target=lnn_target(5), seed_transpiler=2024_10_25_3).layout
        self.assertEqual(
            base.apply_layout(routed),
            base.apply_layout(routed.final_index_layout(filter_ancillas=True)),
        )

        routed_expanded = transpile(qc, target=lnn_target(20), seed_transpiler=2024_10_25_3).layout
        self.assertEqual(
            base.apply_layout(routed_expanded),
            base.apply_layout(
                routed_expanded.final_index_layout(filter_ancillas=True), num_qubits=20
            ),
        )

    def test_apply_layout_none(self):
        self.assertEqual(SparseObservable.zero(0).apply_layout(None), SparseObservable.zero(0))
        self.assertEqual(SparseObservable.zero(0).apply_layout(None, 3), SparseObservable.zero(3))
        self.assertEqual(SparseObservable.zero(5).apply_layout(None), SparseObservable.zero(5))
        self.assertEqual(SparseObservable.zero(3).apply_layout(None, 8), SparseObservable.zero(8))
        self.assertEqual(
            SparseObservable.identity(0).apply_layout(None), SparseObservable.identity(0)
        )
        self.assertEqual(
            SparseObservable.identity(0).apply_layout(None, 8), SparseObservable.identity(8)
        )
        self.assertEqual(
            SparseObservable.identity(2).apply_layout(None), SparseObservable.identity(2)
        )
        self.assertEqual(
            SparseObservable.identity(3).apply_layout(None, 100_000_000),
            SparseObservable.identity(100_000_000),
        )

        terms = [
            ("ZYX", (2, 1, 0), 1j),
            ("", (), -0.5),
            ("+-rl01", (10, 8, 6, 4, 2, 0), 2.0),
        ]
        self.assertEqual(
            SparseObservable.from_sparse_list(terms, num_qubits=12).apply_layout(None),
            SparseObservable.from_sparse_list(terms, num_qubits=12),
        )
        self.assertEqual(
            SparseObservable.from_sparse_list(terms, num_qubits=12).apply_layout(
                None, num_qubits=200
            ),
            SparseObservable.from_sparse_list(terms, num_qubits=200),
        )

    def test_apply_layout_failures(self):
        obs = SparseObservable.from_list([("IIYI", 2.0), ("IIIX", -1j)])
        with self.assertRaisesRegex(ValueError, "duplicate"):
            obs.apply_layout([0, 0, 1, 2])
        with self.assertRaisesRegex(ValueError, "does not account for all contained qubits"):
            obs.apply_layout([0, 1])
        with self.assertRaisesRegex(ValueError, "less than the number of qubits"):
            obs.apply_layout([0, 2, 4, 6])
        with self.assertRaisesRegex(ValueError, "cannot shrink"):
            obs.apply_layout([0, 1], num_qubits=2)
        with self.assertRaisesRegex(ValueError, "cannot shrink"):
            obs.apply_layout(None, num_qubits=2)

        qc = QuantumCircuit(3)
        qc.cx(0, 1)
        qc.cx(1, 2)
        qc.cx(2, 0)
        layout = transpile(qc, target=lnn_target(3), seed_transpiler=2024_10_25).layout
        with self.assertRaisesRegex(ValueError, "cannot shrink"):
            obs.apply_layout(layout, num_qubits=2)

    def test_pauli_bases(self):
        obs = SparseObservable.from_list(
            [
                ("IIIII", 1.0),
                ("IXYZI", 2.0),
                ("+-II+", 1j),
                ("rlrlr", -0.5),
                ("01010", -0.25),
                ("rlYII", 1.0),
            ]
        )
        expected = PauliList(
            [
                Pauli("IIIII"),
                Pauli("IXYZI"),
                Pauli("XXIIX"),
                Pauli("YYYYY"),
                Pauli("ZZZZZ"),
                Pauli("YYYII"),
            ]
        )
        self.assertEqual(obs.pauli_bases(), expected)

    def test_iteration(self):
        self.assertEqual(list(SparseObservable.zero(5)), [])
        self.assertEqual(tuple(SparseObservable.zero(0)), ())

        obs = SparseObservable.from_sparse_list(
            [
                ("Xrl", (4, 2, 1), 2j),
                ("", (), 0.5),
                ("01", (3, 0), -0.25),
                ("+-", (2, 1), 1.0),
                ("YZ", (4, 1), 1j),
            ],
            num_qubits=5,
        )
        bit_term = SparseObservable.BitTerm
        expected = [
            SparseObservable.Term(5, 2j, [bit_term.LEFT, bit_term.RIGHT, bit_term.X], [1, 2, 4]),
            SparseObservable.Term(5, 0.5, [], []),
            SparseObservable.Term(5, -0.25, [bit_term.ONE, bit_term.ZERO], [0, 3]),
            SparseObservable.Term(5, 1.0, [bit_term.MINUS, bit_term.PLUS], [1, 2]),
            SparseObservable.Term(5, 1j, [bit_term.Z, bit_term.Y], [1, 4]),
        ]
        self.assertEqual(list(obs), expected)

    def test_indexing(self):
        obs = SparseObservable.from_sparse_list(
            [
                ("Xrl", (4, 2, 1), 2j),
                ("", (), 0.5),
                ("01", (3, 0), -0.25),
                ("+-", (2, 1), 1.0),
                ("YZ", (4, 1), 1j),
            ],
            num_qubits=5,
        )
        bit_term = SparseObservable.BitTerm
        expected = [
            SparseObservable.Term(5, 2j, [bit_term.LEFT, bit_term.RIGHT, bit_term.X], [1, 2, 4]),
            SparseObservable.Term(5, 0.5, [], []),
            SparseObservable.Term(5, -0.25, [bit_term.ZERO, bit_term.ONE], [3, 0]),
            SparseObservable.Term(5, 1.0, [bit_term.MINUS, bit_term.PLUS], [1, 2]),
            SparseObservable.Term(5, 1j, [bit_term.Y, bit_term.Z], [4, 1]),
        ]
        self.assertEqual(obs[0], expected[0])
        self.assertEqual(obs[-2], expected[-2])
        self.assertEqual(obs[2:4], SparseObservable(expected[2:4]))
        self.assertEqual(obs[1::2], SparseObservable(expected[1::2]))
        self.assertEqual(obs[:], SparseObservable(expected))
        self.assertEqual(obs[-1:-4:-1], SparseObservable(expected[-1:-4:-1]))

    @ddt.data(
        SparseObservable.identity(0),
        SparseObservable.identity(1_000),
        SparseObservable.from_label("IIXIZI"),
        SparseObservable.from_label("X"),
        SparseObservable.from_list([("YIXZII", -0.25)]),
        SparseObservable.from_list([("01rl+-", 0.25 + 0.5j)]),
    )
    def test_term_repr(self, obs):
        # The purpose of this is just to test that the `repr` doesn't crash, rather than asserting
        # that it has any particular form.
        term = obs[0]
        self.assertIsInstance(repr(term), str)
        self.assertIn("SparseObservable.Term", repr(term))

    @ddt.data(
        SparseObservable.identity(0),
        2j * SparseObservable.identity(1),
        SparseObservable.identity(100),
        SparseObservable.from_label("IIX+-rlYZ01IIIII"),
    )
    def test_term_to_observable(self, obs):
        self.assertEqual(obs[0].to_observable(), obs)
        self.assertIsNot(obs[0].to_observable(), obs)

    def test_term_equality(self):
        self.assertEqual(
            SparseObservable.Term(5, 1.0, [], []), SparseObservable.Term(5, 1.0, [], [])
        )
        self.assertNotEqual(
            SparseObservable.Term(5, 1.0, [], []), SparseObservable.Term(8, 1.0, [], [])
        )
        self.assertNotEqual(
            SparseObservable.Term(5, 1.0, [], []), SparseObservable.Term(5, 1j, [], [])
        )
        self.assertNotEqual(
            SparseObservable.Term(5, 1.0, [], []), SparseObservable.Term(8, -1, [], [])
        )

        obs = SparseObservable.from_list(
            [
                ("IIXIZ", 2j),
                ("IIZIX", 2j),
                ("++III", -1.5),
                ("--III", -1.5),
                ("IrIlI", 0.5),
                ("IIrIl", 0.5),
            ]
        )
        self.assertEqual(obs[0], obs[0])
        self.assertEqual(obs[1], obs[1])
        self.assertNotEqual(obs[0], obs[1])
        self.assertEqual(obs[2], obs[2])
        self.assertEqual(obs[3], obs[3])
        self.assertNotEqual(obs[2], obs[3])
        self.assertEqual(obs[4], obs[4])
        self.assertEqual(obs[5], obs[5])
        self.assertNotEqual(obs[4], obs[5])

    @ddt.data(
        SparseObservable.identity(0),
        2j * SparseObservable.identity(1),
        SparseObservable.identity(100),
        SparseObservable.from_label("IIX+-rlYZ01IIIII"),
    )
    def test_term_pickle(self, obs):
        term = obs[0]
        self.assertEqual(pickle.loads(pickle.dumps(term)), term)
        self.assertEqual(copy.copy(term), term)
        self.assertEqual(copy.deepcopy(term), term)

    def test_term_attributes(self):
        term = SparseObservable.from_label("II+IIX0")[0]
        self.assertEqual(term.num_qubits, 7)
        self.assertEqual(term.coeff, 1.0)
        np.testing.assert_equal(
            term.bit_terms,
            np.array(
                [
                    SparseObservable.BitTerm.ZERO,
                    SparseObservable.BitTerm.X,
                    SparseObservable.BitTerm.PLUS,
                ],
                dtype=np.uint8,
            ),
        )
        np.testing.assert_equal(term.indices, np.array([0, 1, 4], dtype=np.uintp))

        term = SparseObservable.identity(10)[0]
        self.assertEqual(term.num_qubits, 10)
        self.assertEqual(term.coeff, 1.0)
        self.assertEqual(list(term.bit_terms), [])
        self.assertEqual(list(term.indices), [])

        term = SparseObservable.from_list([("IIrlZ", 0.5j)])[0]
        self.assertEqual(term.num_qubits, 5)
        self.assertEqual(term.coeff, 0.5j)
        self.assertEqual(
            list(term.bit_terms),
            [
                SparseObservable.BitTerm.Z,
                SparseObservable.BitTerm.LEFT,
                SparseObservable.BitTerm.RIGHT,
            ],
        )
        self.assertEqual(list(term.indices), [0, 1, 2])

    def test_term_new(self):
        expected = SparseObservable.from_label("IIIX+1III")[0]

        self.assertEqual(
            SparseObservable.Term(
                9,
                1.0,
                [
                    SparseObservable.BitTerm.ONE,
                    SparseObservable.BitTerm.PLUS,
                    SparseObservable.BitTerm.X,
                ],
                [3, 4, 5],
            ),
            expected,
        )

        # Constructor should allow being given unsorted inputs, and but them in the right order.
        self.assertEqual(
            SparseObservable.Term(
                9,
                1.0,
                [
                    SparseObservable.BitTerm.PLUS,
                    SparseObservable.BitTerm.X,
                    SparseObservable.BitTerm.ONE,
                ],
                [4, 5, 3],
            ),
            expected,
        )
        self.assertEqual(list(expected.indices), [3, 4, 5])

        with self.assertRaisesRegex(ValueError, "not term-wise increasing"):
            SparseObservable.Term(2, 2j, [SparseObservable.BitTerm.RIGHT] * 2, [0, 0])

    def test_term_pauli_base(self):
        obs = SparseObservable.from_list(
            [
                ("IIIII", 1.0),
                ("IXYZI", 2.0),
                ("+-II+", 1j),
                ("rlrlr", -0.5),
                ("01010", -0.25),
                ("rlYII", 1.0),
            ]
        )
        expected = [
            Pauli("IIIII"),
            Pauli("IXYZI"),
            Pauli("XXIIX"),
            Pauli("YYYYY"),
            Pauli("ZZZZZ"),
            Pauli("YYYII"),
        ]
        self.assertEqual([term.pauli_base() for term in obs], expected)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test library of integer comparison circuits."""

import unittest
import numpy as np
from ddt import ddt, data, unpack

from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library import IntegerComparator
from qiskit.quantum_info import Statevector
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestIntegerComparator(QiskitTestCase):
    """Test the integer comparator circuit."""

    def assertComparisonIsCorrect(self, comp, num_state_qubits, value, geq):
        """Assert that the comparator output is correct."""
        qc = QuantumCircuit(comp.num_qubits)  # initialize circuit
        qc.h(list(range(num_state_qubits)))  # set equal superposition state
        qc.append(comp, list(range(comp.num_qubits)))  # add comparator

        # run simulation
        statevector = Statevector(qc)
        for i, amplitude in enumerate(statevector):
            prob = np.abs(amplitude) ** 2
            if prob > 1e-6:
                # equal superposition
                self.assertEqual(True, np.isclose(1.0, prob * 2.0**num_state_qubits))
                b_value = f"{i:b}".rjust(qc.width(), "0")
                x = int(b_value[(-num_state_qubits):], 2)
                comp_result = int(b_value[-num_state_qubits - 1], 2)
                if geq:
                    self.assertEqual(x >= value, comp_result == 1)
                else:
                    self.assertEqual(x < value, comp_result == 1)

    @data(
        [1, 0, True],
        [1, 1, True],
        [2, -1, True],
        [3, 5, True],
        [3, 2, True],
        [3, 2, False],
        [4, 6, False],
    )
    @unpack
    def test_fixed_value_comparator(self, num_state_qubits, value, geq):
        """Test the fixed value comparator circuit."""
        # build the circuit with the comparator
        comp = IntegerComparator(num_state_qubits, value, geq=geq)
        self.assertComparisonIsCorrect(comp, num_state_qubits, value, geq)

    def test_mutability(self):
        """Test changing the arguments of the comparator."""

        comp = IntegerComparator()

        with self.subTest(msg="missing num state qubits and value"):
            with self.assertRaises(AttributeError):
                _ = str(comp.draw())

        comp.num_state_qubits = 2

        with self.subTest(msg="missing value"):
            with self.assertRaises(AttributeError):
                _ = str(comp.draw())

        comp.value = 0
        comp.geq = True

        with self.subTest(msg="updating num state qubits"):
            comp.num_state_qubits = 1
            self.assertComparisonIsCorrect(comp, 1, 0, True)

        with self.subTest(msg="updating the value"):
            comp.num_state_qubits = 3
            comp.value = 2
            self.assertComparisonIsCorrect(comp, 3, 2, True)

        with self.subTest(msg="updating geq"):
            comp.geq = False
            self.assertComparisonIsCorrect(comp, 3, 2, False)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Assemble function for converting a list of circuits into a qobj"""
import copy
import logging
import uuid
import warnings
from time import time
from typing import Dict, List, Optional, Union

import numpy as np

from qiskit.assembler import assemble_schedules
from qiskit.assembler.run_config import RunConfig
from qiskit.circuit import Parameter, QuantumCircuit, Qubit
from qiskit.exceptions import QiskitError
from qiskit.providers.backend import Backend
from qiskit.pulse import Instruction, LoConfig, Schedule, ScheduleBlock
from qiskit.pulse.channels import PulseChannel
from qiskit.qobj import QasmQobj, PulseQobj, QobjHeader
from qiskit.qobj.utils import MeasLevel, MeasReturnType
from qiskit.utils import deprecate_func
from qiskit.assembler.assemble_circuits import _assemble_circuits

logger = logging.getLogger(__name__)


def _log_assembly_time(start_time, end_time):
    log_msg = f"Total Assembly Time - {((end_time - start_time) * 1000):.5f} (ms)"
    logger.info(log_msg)


# TODO: parallelize over the experiments (serialize each separately, then add global header/config)
@deprecate_func(
    since="1.2",
    removal_timeline="in the 2.0 release",
    additional_msg="The `Qobj` class and related functionality are part of the deprecated "
    "`BackendV1` workflow,  and no longer necessary for `BackendV2`. If a user "
    "workflow requires `Qobj` it likely relies on deprecated functionality and "
    "should be updated to use `BackendV2`.",
)
def assemble(
    experiments: Union[
        QuantumCircuit,
        List[QuantumCircuit],
        Schedule,
        List[Schedule],
        ScheduleBlock,
        List[ScheduleBlock],
    ],
    backend: Optional[Backend] = None,
    qobj_id: Optional[str] = None,
    qobj_header: Optional[Union[QobjHeader, Dict]] = None,
    shots: Optional[int] = None,
    memory: Optional[bool] = False,
    seed_simulator: Optional[int] = None,
    qubit_lo_freq: Optional[List[float]] = None,
    meas_lo_freq: Optional[List[float]] = None,
    qubit_lo_range: Optional[List[float]] = None,
    meas_lo_range: Optional[List[float]] = None,
    schedule_los: Optional[
        Union[
            List[Union[Dict[PulseChannel, float], LoConfig]],
            Union[Dict[PulseChannel, float], LoConfig],
        ]
    ] = None,
    meas_level: Union[int, MeasLevel] = MeasLevel.CLASSIFIED,
    meas_return: Union[str, MeasReturnType] = MeasReturnType.AVERAGE,
    meas_map: Optional[List[List[Qubit]]] = None,
    memory_slot_size: int = 100,
    rep_time: Optional[int] = None,
    rep_delay: Optional[float] = None,
    parameter_binds: Optional[List[Dict[Parameter, float]]] = None,
    parametric_pulses: Optional[List[str]] = None,
    init_qubits: bool = True,
    **run_config: Dict,
) -> Union[QasmQobj, PulseQobj]:
    """Assemble a list of circuits or pulse schedules into a ``Qobj``.

    This function serializes the payloads, which could be either circuits or schedules,
    to create ``Qobj`` "experiments". It further annotates the experiment payload with
    header and configurations.

    NOTE: ``Backend.options`` is not used within assemble. The required values
    (previously given by backend.set_options) should be manually extracted
    from options and supplied directly when calling.

    Args:
        experiments: Circuit(s) or pulse schedule(s) to execute
        backend: If set, some runtime options are automatically grabbed from
            ``backend.configuration()`` and ``backend.defaults()``.
            If any other option is explicitly set (e.g., ``rep_time``), it
            will override the backend's.
            If any other options is set in the run_config, it will
            also override the backend's.
        qobj_id: String identifier to annotate the ``Qobj``
        qobj_header: User input that will be inserted in ``Qobj`` header, and will also be
            copied to the corresponding Result header. Headers do not affect the run.
        shots: Number of repetitions of each circuit, for sampling. Default: 1024
            or ``max_shots`` from the backend configuration, whichever is smaller
        memory: If ``True``, per-shot measurement bitstrings are returned as well
            (provided the backend supports it). For OpenPulse jobs, only
            measurement level 2 supports this option.
        seed_simulator: Random seed to control sampling, for when backend is a simulator
        qubit_lo_freq: List of job level qubit drive LO frequencies in Hz. Overridden by
            ``schedule_los`` if specified. Must have length ``n_qubits.``
        meas_lo_freq: List of measurement LO frequencies in Hz. Overridden by ``schedule_los`` if
            specified. Must have length ``n_qubits.``
        qubit_lo_range: List of job level drive LO ranges each of form ``[range_min, range_max]``
            in Hz. Used to validate ``qubit_lo_freq``. Must have length ``n_qubits.``
        meas_lo_range: List of job level measurement LO ranges each of form
            ``[range_min, range_max]`` in Hz. Used to validate ``meas_lo_freq``. Must have length
            ``n_qubits.``
        schedule_los: Experiment level (ie circuit or schedule) LO frequency configurations for
            qubit drive and measurement channels. These values override the job level values from
            ``default_qubit_los`` and ``default_meas_los``. Frequencies are in Hz. Settable for
            OpenQASM 2 and pulse jobs.
        meas_level: Set the appropriate level of the measurement output for pulse experiments.
        meas_return: Level of measurement data for the backend to return.

            For ``meas_level`` 0 and 1:
                * ``single`` returns information from every shot.
                * ``avg`` returns average measurement output (averaged over number of shots).
        meas_map: List of lists, containing qubits that must be measured together.
        memory_slot_size: Size of each memory slot if the output is Level 0.
        rep_time (int): Time per program execution in seconds. Must be from the list provided
            by the backend (``backend.configuration().rep_times``). Defaults to the first entry.
        rep_delay (float): Delay between programs in seconds. Only supported on certain
            backends (if ``backend.configuration().dynamic_reprate_enabled=True``). If supported,
            ``rep_delay`` will be used instead of ``rep_time`` and must be from the range supplied
            by the backend (``backend.configuration().rep_delay_range``). Default is given by
            ``backend.configuration().default_rep_delay``.
        parameter_binds: List of Parameter bindings over which the set of experiments will be
            executed. Each list element (bind) should be of the form
            {Parameter1: value1, Parameter2: value2, ...}. All binds will be
            executed across all experiments; e.g., if parameter_binds is a
            length-n list, and there are m experiments, a total of m x n
            experiments will be run (one for each experiment/bind pair).
        parametric_pulses: A list of pulse shapes which are supported internally on the backend.
            Example::

            ['gaussian', 'constant']
        init_qubits: Whether to reset the qubits to the ground state for each shot.
                     Default: ``True``.
        **run_config: Extra arguments used to configure the run (e.g., for Aer configurable
            backends). Refer to the backend documentation for details on these
            arguments.

    Returns:
            A ``Qobj`` that can be run on a backend. Depending on the type of input,
            this will be either a ``QasmQobj`` or a ``PulseQobj``.

    Raises:
        QiskitError: if the input cannot be interpreted as either circuits or schedules
    """
    return _assemble(
        experiments,
        backend,
        qobj_id,
        qobj_header,
        shots,
        memory,
        seed_simulator,
        qubit_lo_freq,
        meas_lo_freq,
        qubit_lo_range,
        meas_lo_range,
        schedule_los,
        meas_level,
        meas_return,
        meas_map,
        memory_slot_size,
        rep_time,
        rep_delay,
        parameter_binds,
        parametric_pulses,
        init_qubits,
        **run_config,
    )


# Note for future: this method is used in `BasicSimulator` and may need to be kept past the
# `assemble` removal deadline (2.0). If it is kept (potentially in a different location),
# we will need an alternative for the backend.configuration() access that currently takes
# place in L566 (`parse_circuit_args`) and L351 (`parse_common_args`)
# because backend.configuration() is also set for removal in 2.0.
# The ultimate goal will be to move away from relying on any kind of `assemble` implementation
# because of how tightly coupled it is to these legacy data structures. But as a transition step,
# given that we would only have to support the subcase of `BasicSimulator`, we could probably just
# inline the relevant config values that are already hardcoded in the basic simulator configuration
# generator.
def _assemble(
    experiments: Union[
        QuantumCircuit,
        List[QuantumCircuit],
        Schedule,
        List[Schedule],
        ScheduleBlock,
        List[ScheduleBlock],
    ],
    backend: Optional[Backend] = None,
    qobj_id: Optional[str] = None,
    qobj_header: Optional[Union[QobjHeader, Dict]] = None,
    shots: Optional[int] = None,
    memory: Optional[bool] = False,
    seed_simulator: Optional[int] = None,
    qubit_lo_freq: Optional[List[float]] = None,
    meas_lo_freq: Optional[List[float]] = None,
    qubit_lo_range: Optional[List[float]] = None,
    meas_lo_range: Optional[List[float]] = None,
    schedule_los: Optional[
        Union[
            List[Union[Dict[PulseChannel, float], LoConfig]],
            Union[Dict[PulseChannel, float], LoConfig],
        ]
    ] = None,
    meas_level: Union[int, MeasLevel] = MeasLevel.CLASSIFIED,
    meas_return: Union[str, MeasReturnType] = MeasReturnType.AVERAGE,
    meas_map: Optional[List[List[Qubit]]] = None,
    memory_slot_size: int = 100,
    rep_time: Optional[int] = None,
    rep_delay: Optional[float] = None,
    parameter_binds: Optional[List[Dict[Parameter, float]]] = None,
    parametric_pulses: Optional[List[str]] = None,
    init_qubits: bool = True,
    **run_config: Dict,
) -> Union[QasmQobj, PulseQobj]:
    start_time = time()
    experiments = experiments if isinstance(experiments, list) else [experiments]
    pulse_qobj = any(isinstance(exp, (ScheduleBlock, Schedule, Instruction)) for exp in experiments)
    with warnings.catch_warnings():
        # The Qobj class is deprecated, the backend.configuration() method is too
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        qobj_id, qobj_header, run_config_common_dict = _parse_common_args(
            backend,
            qobj_id,
            qobj_header,
            shots,
            memory,
            seed_simulator,
            init_qubits,
            rep_delay,
            qubit_lo_freq,
            meas_lo_freq,
            qubit_lo_range,
            meas_lo_range,
            schedule_los,
            pulse_qobj=pulse_qobj,
            **run_config,
        )

    # assemble either circuits or schedules
    if all(isinstance(exp, QuantumCircuit) for exp in experiments):
        with warnings.catch_warnings():
            # Internally calls deprecated BasicSimulator.configuration()`
            warnings.filterwarnings(
                "ignore",
                category=DeprecationWarning,
                message=r".+\.basic_provider\.basic_simulator\.BasicSimulator\.configuration.+",
                module="qiskit",
            )
            run_config = _parse_circuit_args(
                parameter_binds,
                backend,
                meas_level,
                meas_return,
                parametric_pulses,
                **run_config_common_dict,
            )

        # If circuits are parameterized, bind parameters and remove from run_config
        bound_experiments, run_config = _expand_parameters(
            circuits=experiments, run_config=run_config
        )
        end_time = time()
        _log_assembly_time(start_time, end_time)
        return _assemble_circuits(
            circuits=bound_experiments,
            qobj_id=qobj_id,
            qobj_header=qobj_header,
            run_config=run_config,
        )

    elif all(isinstance(exp, (ScheduleBlock, Schedule, Instruction)) for exp in experiments):
        run_config = _parse_pulse_args(
            backend,
            meas_level,
            meas_return,
            meas_map,
            memory_slot_size,
            rep_time,
            parametric_pulses,
            **run_config_common_dict,
        )

        end_time = time()
        _log_assembly_time(start_time, end_time)
        return assemble_schedules(
            schedules=experiments, qobj_id=qobj_id, qobj_header=qobj_header, run_config=run_config
        )

    else:
        raise QiskitError("bad input to assemble() function; must be either circuits or schedules")


# TODO: rework to return a list of RunConfigs (one for each experiments), and a global one
def _parse_common_args(
    backend,
    qobj_id,
    qobj_header,
    shots,
    memory,
    seed_simulator,
    init_qubits,
    rep_delay,
    qubit_lo_freq,
    meas_lo_freq,
    qubit_lo_range,
    meas_lo_range,
    schedule_los,
    pulse_qobj=False,
    **run_config,
):
    """Resolve the various types of args allowed to the assemble() function through
    duck typing, overriding args, etc. Refer to the assemble() docstring for details on
    what types of inputs are allowed.

    Here the args are resolved by converting them to standard instances, and prioritizing
    them in case a run option is passed through multiple args (explicitly setting an arg
    has more priority than the arg set by backend)

    Returns:
        RunConfig: a run config, which is a standardized object that configures the qobj
            and determines the runtime environment.

    Raises:
        QiskitError:
            - If the memory arg is True and the backend does not support memory.
            - If ``shots`` exceeds ``max_shots`` for the configured backend.
            - If ``shots`` are not int type.
            - If any of qubit or meas lo's, or associated ranges do not have length equal to
            ``n_qubits``.
            - If qubit or meas lo's do not fit into prescribed ranges.
    """
    # grab relevant info from backend if it exists
    backend_config = None
    backend_defaults = None
    n_qubits = None
    if backend:
        backend_config = backend.configuration()
        n_qubits = backend_config.n_qubits
        # check for memory flag applied to backend that does not support memory
        if memory and not backend_config.memory:
            raise QiskitError(f"memory not supported by backend {backend_config.backend_name}")

        # try to set defaults for pulse, other leave as None
        pulse_param_set = (
            qubit_lo_freq is not None
            or meas_lo_freq is not None
            or qubit_lo_range is not None
            or meas_lo_range is not None
            or schedule_los is not None
        )
        if pulse_qobj or (backend_config.open_pulse and pulse_param_set):
            try:
                backend_defaults = backend.defaults()
            except AttributeError:
                pass

    # an identifier for the Qobj
    qobj_id = qobj_id or str(uuid.uuid4())

    # The header that goes at the top of the Qobj (and later Result)
    # we process it as dict, then write entries that are not None to a QobjHeader object
    qobj_header = qobj_header or {}
    if isinstance(qobj_header, QobjHeader):
        qobj_header = qobj_header.to_dict()
    backend_name = getattr(backend_config, "backend_name", None)
    backend_version = getattr(backend_config, "backend_version", None)
    qobj_header = {
        "backend_name": backend_name,
        "backend_version": backend_version,
        **qobj_header,
    }
    qobj_header = QobjHeader(**{k: v for k, v in qobj_header.items() if v is not None})

    max_shots = getattr(backend_config, "max_shots", None)
    if shots is None:
        if max_shots:
            shots = min(1024, max_shots)
        else:
            shots = 1024
    elif not isinstance(shots, (int, np.integer)):
        raise QiskitError("Argument 'shots' should be of type 'int'")
    elif max_shots and max_shots < shots:
        raise QiskitError(
            f"Number of shots specified: {max_shots} exceeds max_shots property of the "
            f"backend: {max_shots}."
        )

    dynamic_reprate_enabled = getattr(backend_config, "dynamic_reprate_enabled", False)
    if dynamic_reprate_enabled:
        default_rep_delay = getattr(backend_config, "default_rep_delay", None)
        rep_delay_range = getattr(backend_config, "rep_delay_range", None)
        rep_delay = _parse_rep_delay(rep_delay, default_rep_delay, rep_delay_range)
    else:
        if rep_delay is not None:
            rep_delay = None
            warnings.warn(
                "Dynamic rep rates not supported on this backend, cannot use rep_delay.",
                RuntimeWarning,
            )

    qubit_lo_freq = qubit_lo_freq or getattr(backend_defaults, "qubit_freq_est", None)
    meas_lo_freq = meas_lo_freq or getattr(backend_defaults, "meas_freq_est", None)

    qubit_lo_range = qubit_lo_range or getattr(backend_config, "qubit_lo_range", None)
    meas_lo_range = meas_lo_range or getattr(backend_config, "meas_lo_range", None)

    # check that LO frequencies are in the perscribed range
    _check_lo_freqs(qubit_lo_freq, qubit_lo_range, "qubit")
    _check_lo_freqs(meas_lo_freq, meas_lo_range, "meas")

    # configure experiment level LO frequencies
    schedule_los = schedule_los or []
    if isinstance(schedule_los, (LoConfig, dict)):
        schedule_los = [schedule_los]

    # Convert to LoConfig if LO configuration supplied as dictionary
    schedule_los = [
        lo_config if isinstance(lo_config, LoConfig) else LoConfig(lo_config)
        for lo_config in schedule_los
    ]

    # create run configuration and populate
    run_config_dict = {
        "shots": shots,
        "memory": memory,
        "seed_simulator": seed_simulator,
        "init_qubits": init_qubits,
        "rep_delay": rep_delay,
        "qubit_lo_freq": qubit_lo_freq,
        "meas_lo_freq": meas_lo_freq,
        "qubit_lo_range": qubit_lo_range,
        "meas_lo_range": meas_lo_range,
        "schedule_los": schedule_los,
        "n_qubits": n_qubits,
        **run_config,
    }

    return qobj_id, qobj_header, run_config_dict


def _check_lo_freqs(
    lo_freq: Union[List[float], None],
    lo_range: Union[List[float], None],
    lo_type: str,
):
    """Check that LO frequencies are within the perscribed LO range.

    NOTE: Only checks if frequency/range lists have equal length. And does not check that the lists
    have length ``n_qubits``. This is because some backends, like simulator backends, do not
    require these constraints. For real hardware, these parameters will be validated on the backend.

    Args:
        lo_freq: List of LO frequencies.
        lo_range: Nested list of LO frequency ranges. Inner list is of the form
            ``[lo_min, lo_max]``.
        lo_type: The type of LO value--"qubit" or "meas".

    Raises:
        QiskitError:
            - If each element of the LO range is not a 2d list.
            - If the LO frequency is not in the LO range for a given qubit.
    """
    if lo_freq and lo_range and len(lo_freq) == len(lo_range):
        for i, freq in enumerate(lo_freq):
            freq_range = lo_range[i]
            if not (isinstance(freq_range, list) and len(freq_range) == 2):
                raise QiskitError(f"Each element of {lo_type} LO range must be a 2d list.")
            if freq < freq_range[0] or freq > freq_range[1]:
                raise QiskitError(
                    f"Qubit {i} {lo_type} LO frequency is {freq}. "
                    f"The range is [{freq_range[0]}, {freq_range[1]}]."
                )


def _parse_pulse_args(
    backend,
    meas_level,
    meas_return,
    meas_map,
    memory_slot_size,
    rep_time,
    parametric_pulses,
    **run_config,
):
    """Build a pulse RunConfig replacing unset arguments with defaults derived from the `backend`.
    See `assemble` for more information on the required arguments.

    Returns:
        RunConfig: a run config, which is a standardized object that configures the qobj
            and determines the runtime environment.
    Raises:
        QiskitError: If the given meas_level is not allowed for the given `backend`.
    """
    # grab relevant info from backend if it exists
    backend_config = None
    if backend:
        backend_config = backend.configuration()

        if meas_level not in getattr(backend_config, "meas_levels", [MeasLevel.CLASSIFIED]):
            raise QiskitError(
                f"meas_level = {meas_level} not supported for backend "
                f"{backend_config.backend_name}, only {backend_config.meas_levels} is supported"
            )

    meas_map = meas_map or getattr(backend_config, "meas_map", None)
    dynamic_reprate_enabled = getattr(backend_config, "dynamic_reprate_enabled", False)

    rep_time = rep_time or getattr(backend_config, "rep_times", None)
    if rep_time:
        if dynamic_reprate_enabled:
            warnings.warn(
                "Dynamic rep rates are supported on this backend. 'rep_delay' will be "
                "used instead of 'rep_time'.",
                RuntimeWarning,
            )
        if isinstance(rep_time, list):
            rep_time = rep_time[0]
        rep_time = int(rep_time * 1e6)  # convert sec to μs
    if parametric_pulses is None:
        parametric_pulses = getattr(backend_config, "parametric_pulses", [])

    # create run configuration and populate
    run_config_dict = {
        "meas_level": meas_level,
        "meas_return": meas_return,
        "meas_map": meas_map,
        "memory_slot_size": memory_slot_size,
        "rep_time": rep_time,
        "parametric_pulses": parametric_pulses,
        **run_config,
    }
    run_config = RunConfig(**{k: v for k, v in run_config_dict.items() if v is not None})

    return run_config


def _parse_circuit_args(
    parameter_binds, backend, meas_level, meas_return, parametric_pulses, **run_config
):
    """Build a circuit RunConfig replacing unset arguments with defaults derived from the `backend`.
    See `assemble` for more information on the required arguments.

    Returns:
        RunConfig: a run config, which is a standardized object that configures the qobj
            and determines the runtime environment.
    """
    parameter_binds = parameter_binds or []
    # create run configuration and populate
    run_config_dict = {"parameter_binds": parameter_binds, **run_config}
    if parametric_pulses is None:
        if backend:
            with warnings.catch_warnings():
                # TODO (2.0): See comment on L192 regarding backend.configuration removal
                warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
                run_config_dict["parametric_pulses"] = getattr(
                    backend.configuration(), "parametric_pulses", []
                )
        else:
            run_config_dict["parametric_pulses"] = []
    else:
        run_config_dict["parametric_pulses"] = parametric_pulses
    if meas_level:
        run_config_dict["meas_level"] = meas_level
        # only enable `meas_return` if `meas_level` isn't classified
        if meas_level != MeasLevel.CLASSIFIED:
            run_config_dict["meas_return"] = meas_return

    run_config = RunConfig(**{k: v for k, v in run_config_dict.items() if v is not None})

    return run_config


def _parse_rep_delay(
    rep_delay: float, default_rep_delay: float, rep_delay_range: List[float]
) -> float:
    """Parse and set ``rep_delay`` parameter in runtime config.

    Args:
        rep_delay: Initial rep delay.
        default_rep_delay: Backend default rep delay.
        rep_delay_range: Backend list defining allowable range of rep delays.

    Raises:
        QiskitError: If rep_delay is not in the backend rep_delay_range.
    Returns:
        float: Modified rep delay after parsing.
    """
    if rep_delay is None:
        rep_delay = default_rep_delay

    if rep_delay is not None:
        # check that rep_delay is in rep_delay_range
        if rep_delay_range is not None and isinstance(rep_delay_range, list):
            if len(rep_delay_range) != 2:
                raise QiskitError(
                    f"Backend rep_delay_range {rep_delay_range} must be a list with two entries."
                )
            if not rep_delay_range[0] <= rep_delay <= rep_delay_range[1]:
                raise QiskitError(
                    f"Supplied rep delay {rep_delay} not in the supported "
                    f"backend range {rep_delay_range}"
                )
        rep_delay = rep_delay * 1e6  # convert sec to μs

    return rep_delay


def _expand_parameters(circuits, run_config):
    """Verifies that there is a single common set of parameters shared between
    all circuits and all parameter binds in the run_config. Returns an expanded
    list of circuits (if parameterized) with all parameters bound, and a copy of
    the run_config with parameter_binds cleared.

    If neither the circuits nor the run_config specify parameters, the two are
    returned unmodified.

    Raises:
        QiskitError: if run_config parameters are not compatible with circuit parameters

    Returns:
        Tuple(List[QuantumCircuit], RunConfig):
          - List of input circuits expanded and with parameters bound
          - RunConfig with parameter_binds removed
    """
    parameter_binds = run_config.parameter_binds

    if parameter_binds and any(parameter_binds) or any(circuit.parameters for circuit in circuits):

        # Unroll params here in order to handle ParamVects
        all_bind_parameters = [
            QuantumCircuit()._unroll_param_dict(bind).keys() for bind in parameter_binds
        ]

        all_circuit_parameters = [circuit.parameters for circuit in circuits]

        # Collect set of all unique parameters across all circuits and binds
        unique_parameters = {
            param
            for param_list in all_bind_parameters + all_circuit_parameters
            for param in param_list
        }

        # Check that all parameters are common to all circuits and binds
        if (
            not all_bind_parameters
            or not all_circuit_parameters
            or any(unique_parameters != bind_params for bind_params in all_bind_parameters)
            or any(unique_parameters != parameters for parameters in all_circuit_parameters)
        ):
            raise QiskitError(
                (
                    "Mismatch between run_config.parameter_binds and all circuit parameters. "
                    + "Parameter binds: {} "
                    + "Circuit parameters: {}"
                ).format(all_bind_parameters, all_circuit_parameters)
            )

        circuits = [
            circuit.assign_parameters(binds) for circuit in circuits for binds in parameter_binds
        ]

        # All parameters have been expanded and bound, so remove from run_config
        run_config = copy.deepcopy(run_config)
        run_config.parameter_binds = []

    return circuits, run_config



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A swap strategy pass for blocks of commuting gates."""
from __future__ import annotations
from collections import defaultdict

from qiskit.circuit import Gate, QuantumCircuit, Qubit
from qiskit.converters import circuit_to_dag
from qiskit.dagcircuit import DAGCircuit, DAGOpNode
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.transpiler.layout import Layout
from qiskit.transpiler.passes.routing.commuting_2q_gate_routing.swap_strategy import SwapStrategy
from qiskit.transpiler.passes.routing.commuting_2q_gate_routing.commuting_2q_block import (
    Commuting2qBlock,
)


class Commuting2qGateRouter(TransformationPass):
    """A class to swap route one or more commuting gates to the coupling map.

    This pass routes blocks of commuting two-qubit gates encapsulated as
    :class:`.Commuting2qBlock` instructions. This pass will not apply to other instructions.
    The mapping to the coupling map is done using swap strategies, see :class:`.SwapStrategy`.
    The swap strategy should suit the problem and the coupling map. This transpiler pass
    should ideally be executed before the quantum circuit is enlarged with any idle ancilla
    qubits. Otherwise, we may swap qubits outside the portion of the chip we want to use.
    Therefore, the swap strategy and its associated coupling map do not represent physical
    qubits. Instead, they represent an intermediate mapping that corresponds to the physical
    qubits once the initial layout is applied. The example below shows how to map a four
    qubit :class:`.PauliEvolutionGate` to qubits 0, 1, 3, and 4 of the five qubit device with
    the coupling map

    .. code-block:: text

        0 -- 1 -- 2
             |
             3
             |
             4

    To do this we use a line swap strategy for qubits 0, 1, 3, and 4 defined it in terms
    of virtual qubits 0, 1, 2, and 3.

    .. code-block:: python

        from qiskit import QuantumCircuit
        from qiskit.circuit.library import PauliEvolutionGate
        from qiskit.quantum_info import SparsePauliOp
        from qiskit.transpiler import Layout, CouplingMap, PassManager
        from qiskit.transpiler.passes import FullAncillaAllocation
        from qiskit.transpiler.passes import EnlargeWithAncilla
        from qiskit.transpiler.passes import ApplyLayout
        from qiskit.transpiler.passes import SetLayout

        from qiskit.transpiler.passes.routing.commuting_2q_gate_routing import (
            SwapStrategy,
            FindCommutingPauliEvolutions,
            Commuting2qGateRouter,
        )

        # Define the circuit on virtual qubits
        op = SparsePauliOp.from_list([("IZZI", 1), ("ZIIZ", 2), ("ZIZI", 3)])
        circ = QuantumCircuit(4)
        circ.append(PauliEvolutionGate(op, 1), range(4))

        # Define the swap strategy on qubits before the initial_layout is applied.
        swap_strat = SwapStrategy.from_line([0, 1, 2, 3])

        # Chose qubits 0, 1, 3, and 4 from the backend coupling map shown above.
        backend_cmap = CouplingMap(couplinglist=[(0, 1), (1, 2), (1, 3), (3, 4)])
        initial_layout = Layout.from_intlist([0, 1, 3, 4], *circ.qregs)

        pm_pre = PassManager(
            [
                FindCommutingPauliEvolutions(),
                Commuting2qGateRouter(swap_strat),
                SetLayout(initial_layout),
                FullAncillaAllocation(backend_cmap),
                EnlargeWithAncilla(),
                ApplyLayout(),
            ]
        )

        # Insert swap gates, map to initial_layout and finally enlarge with ancilla.
        pm_pre.run(circ).draw("mpl")

    This pass manager relies on the ``current_layout`` which corresponds to the qubit layout as
    swap gates are applied. The pass will traverse all nodes in the dag. If a node should be
    routed using a swap strategy then it will be decomposed into sub-instructions with swap
    layers in between and the ``current_layout`` will be modified. Nodes that should not be
    routed using swap strategies will be added back to the dag taking the ``current_layout``
    into account.
    """

    def __init__(
        self,
        swap_strategy: SwapStrategy | None = None,
        edge_coloring: dict[tuple[int, int], int] | None = None,
    ) -> None:
        r"""
        Args:
            swap_strategy: An instance of a :class:`.SwapStrategy` that holds the swap layers
                that are used, and the order in which to apply them, to map the instruction to
                the hardware. If this field is not given, it should be contained in the
                property set of the pass. This allows other passes to determine the most
                appropriate swap strategy at run-time.
            edge_coloring: An optional edge coloring of the coupling map (I.e. no two edges that
                share a node have the same color). If the edge coloring is given then the commuting
                gates that can be simultaneously applied given the current qubit permutation are
                grouped according to the edge coloring and applied according to this edge
                coloring. Here, a color is an int which is used as the index to define and
                access the groups of commuting gates that can be applied simultaneously.
                If the edge coloring is not given then the sets will be built-up using a
                greedy algorithm. The edge coloring is useful to position gates such as
                ``RZZGate``\s next to swap gates to exploit CX cancellations.
        """
        super().__init__()
        self._swap_strategy = swap_strategy
        self._bit_indices: dict[Qubit, int] | None = None
        self._edge_coloring = edge_coloring

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        """Run the pass by decomposing the nodes it applies on.

        Args:
            dag: The dag to which we will add swaps.

        Returns:
            A dag where swaps have been added for the intended gate type.

        Raises:
            TranspilerError: If the swap strategy was not given at init time and there is
                no swap strategy in the property set.
            TranspilerError: If the quantum circuit contains more than one qubit register.
            TranspilerError: If there are qubits that are not contained in the quantum register.
        """
        if self._swap_strategy is None:
            swap_strategy = self.property_set["swap_strategy"]

            if swap_strategy is None:
                raise TranspilerError("No swap strategy given at init or in the property set.")
        else:
            swap_strategy = self._swap_strategy

        if len(dag.qregs) != 1:
            raise TranspilerError(
                f"{self.__class__.__name__} runs on circuits with one quantum register."
            )

        if len(dag.qubits) != next(iter(dag.qregs.values())).size:
            raise TranspilerError("Circuit has qubits not contained in the qubit register.")

        # Fix output permutation -- copied from ElidePermutations
        input_qubit_mapping = {qubit: index for index, qubit in enumerate(dag.qubits)}
        self.property_set["original_layout"] = Layout(input_qubit_mapping)
        if self.property_set["original_qubit_indices"] is None:
            self.property_set["original_qubit_indices"] = input_qubit_mapping

        new_dag = dag.copy_empty_like()
        current_layout = Layout.generate_trivial_layout(*dag.qregs.values())

        # Used to keep track of nodes that do not decompose using swap strategies.
        accumulator = new_dag.copy_empty_like()

        for node in dag.topological_op_nodes():
            if isinstance(node.op, Commuting2qBlock):

                # Check that the swap strategy creates enough connectivity for the node.
                self._check_edges(dag, node, swap_strategy)

                # Compose any accumulated non-swap strategy gates to the dag
                accumulator = self._compose_non_swap_nodes(accumulator, current_layout, new_dag)

                # Decompose the swap-strategy node and add to the dag.
                new_dag.compose(self.swap_decompose(dag, node, current_layout, swap_strategy))
            else:
                accumulator.apply_operation_back(node.op, node.qargs, node.cargs)

        self._compose_non_swap_nodes(accumulator, current_layout, new_dag)

        self.property_set["virtual_permutation_layout"] = current_layout

        return new_dag

    def _compose_non_swap_nodes(
        self, accumulator: DAGCircuit, layout: Layout, new_dag: DAGCircuit
    ) -> DAGCircuit:
        """Add all the non-swap strategy nodes that we have accumulated up to now.

        This method also resets the node accumulator to an empty dag.

        Args:
            layout: The current layout that keeps track of the swaps.
            new_dag: The new dag that we are building up.
            accumulator: A DAG to keep track of nodes that do not decompose
                using swap strategies.

        Returns:
            A new accumulator with the same registers as ``new_dag``.
        """
        # Add all the non-swap strategy nodes that we have accumulated up to now.
        order = layout.reorder_bits(new_dag.qubits)
        order_bits: list[int | None] = [None] * len(layout)
        for idx, val in enumerate(order):
            order_bits[val] = idx

        new_dag.compose(accumulator, qubits=order_bits)

        # Re-initialize the node accumulator
        return new_dag.copy_empty_like()

    def _position_in_cmap(self, dag: DAGCircuit, j: int, k: int, layout: Layout) -> tuple[int, ...]:
        """A helper function to track the movement of virtual qubits through the swaps.

        Args:
            j: The index of decision variable j (i.e. virtual qubit).
            k: The index of decision variable k (i.e. virtual qubit).
            layout: The current layout that takes into account previous swap gates.

        Returns:
            The position in the coupling map of the virtual qubits j and k as a tuple.
        """
        bit0 = dag.find_bit(layout.get_physical_bits()[j]).index
        bit1 = dag.find_bit(layout.get_physical_bits()[k]).index

        return bit0, bit1

    def _build_sub_layers(
        self, current_layer: dict[tuple[int, int], Gate]
    ) -> list[dict[tuple[int, int], Gate]]:
        """A helper method to build-up sets of gates to simultaneously apply.

        This is done with an edge coloring if the ``edge_coloring`` init argument was given or with
        a greedy algorithm if not. With an edge coloring all gates on edges with the same color
        will be applied simultaneously. These sublayers are applied in the order of their color,
        which is an int, in increasing color order.

        Args:
            current_layer: All gates in the current layer can be applied given the qubit ordering
                of the current layout. However, not all gates in the current layer can be applied
                simultaneously. This function creates sub-layers by building up sub-layers
                of gates. All gates in a sub-layer can simultaneously be applied given the coupling
                map and current qubit configuration.

        Returns:
             A list of gate dicts that can be applied. The gates a position 0 are applied first.
             A gate dict has the qubit tuple as key and the gate to apply as value.
        """
        if self._edge_coloring is not None:
            return self._edge_coloring_build_sub_layers(current_layer)
        else:
            return self._greedy_build_sub_layers(current_layer)

    def _edge_coloring_build_sub_layers(
        self, current_layer: dict[tuple[int, int], Gate]
    ) -> list[dict[tuple[int, int], Gate]]:
        """The edge coloring method of building sub-layers of commuting gates."""
        sub_layers: list[dict[tuple[int, int], Gate]] = [
            {} for _ in set(self._edge_coloring.values())
        ]
        for edge, gate in current_layer.items():
            color = self._edge_coloring[edge]
            sub_layers[color][edge] = gate

        return sub_layers

    @staticmethod
    def _greedy_build_sub_layers(
        current_layer: dict[tuple[int, int], Gate]
    ) -> list[dict[tuple[int, int], Gate]]:
        """The greedy method of building sub-layers of commuting gates."""
        sub_layers = []
        while len(current_layer) > 0:
            current_sub_layer, remaining_gates = {}, {}
            blocked_vertices: set[tuple] = set()

            for edge, evo_gate in current_layer.items():
                if blocked_vertices.isdisjoint(edge):
                    current_sub_layer[edge] = evo_gate

                    # A vertex becomes blocked once a gate is applied to it.
                    blocked_vertices = blocked_vertices.union(edge)
                else:
                    remaining_gates[edge] = evo_gate

            current_layer = remaining_gates
            sub_layers.append(current_sub_layer)

        return sub_layers

    def swap_decompose(
        self, dag: DAGCircuit, node: DAGOpNode, current_layout: Layout, swap_strategy: SwapStrategy
    ) -> DAGCircuit:
        """Take an instance of :class:`.Commuting2qBlock` and map it to the coupling map.

        The mapping is done with the swap strategy.

        Args:
            dag: The dag which contains the :class:`.Commuting2qBlock` we route.
            node: A node whose operation is a :class:`.Commuting2qBlock`.
            current_layout: The layout before the swaps are applied. This function will
                modify the layout so that subsequent gates can be properly composed on the dag.
            swap_strategy: The swap strategy used to decompose the node.

        Returns:
            A dag that is compatible with the coupling map where swap gates have been added
            to map the gates in the :class:`.Commuting2qBlock` to the hardware.
        """
        trivial_layout = Layout.generate_trivial_layout(*dag.qregs.values())
        gate_layers = self._make_op_layers(dag, node.op, current_layout, swap_strategy)

        # Iterate over and apply gate layers
        max_distance = max(gate_layers.keys())

        circuit_with_swap = QuantumCircuit(len(dag.qubits))

        for i in range(max_distance + 1):
            # Get current layer and replace the problem indices j,k by the corresponding
            # positions in the coupling map. The current layer corresponds
            # to all the gates that can be applied at the ith swap layer.
            current_layer = {}
            for (j, k), local_gate in gate_layers.get(i, {}).items():
                current_layer[self._position_in_cmap(dag, j, k, current_layout)] = local_gate

            # Not all gates that are applied at the ith swap layer can be applied at the same
            # time. We therefore greedily build sub-layers.
            sub_layers = self._build_sub_layers(current_layer)

            # Apply sub-layers
            for sublayer in sub_layers:
                for edge, local_gate in sublayer.items():
                    circuit_with_swap.append(local_gate, edge)

            # Apply SWAP gates
            if i < max_distance:
                for swap in swap_strategy.swap_layer(i):
                    (j, k) = [trivial_layout.get_physical_bits()[vertex] for vertex in swap]

                    circuit_with_swap.swap(j, k)
                    current_layout.swap(j, k)

        return circuit_to_dag(circuit_with_swap)

    def _make_op_layers(
        self, dag: DAGCircuit, op: Commuting2qBlock, layout: Layout, swap_strategy: SwapStrategy
    ) -> dict[int, dict[tuple, Gate]]:
        """Creates layers of two-qubit gates based on the distance in the swap strategy."""

        gate_layers: dict[int, dict[tuple, Gate]] = defaultdict(dict)

        for node in op.node_block:
            edge = (dag.find_bit(node.qargs[0]).index, dag.find_bit(node.qargs[1]).index)

            bit0 = layout.get_virtual_bits()[dag.qubits[edge[0]]]
            bit1 = layout.get_virtual_bits()[dag.qubits[edge[1]]]

            distance = swap_strategy.distance_matrix[bit0, bit1]

            gate_layers[distance][edge] = node.op

        return gate_layers

    def _check_edges(self, dag: DAGCircuit, node: DAGOpNode, swap_strategy: SwapStrategy):
        """Check if the swap strategy can create the required connectivity.

        Args:
            node: The dag node for which to check if the swap strategy provides enough connectivity.
            swap_strategy: The swap strategy that is being used.

        Raises:
            TranspilerError: If there is an edge that the swap strategy cannot accommodate
                and if the pass has been configured to raise on such issues.
        """
        required_edges = set()

        for sub_node in node.op:
            edge = (dag.find_bit(sub_node.qargs[0]).index, dag.find_bit(sub_node.qargs[1]).index)
            required_edges.add(edge)

        # Check that the swap strategy supports all required edges
        if not required_edges.issubset(swap_strategy.possible_edges):
            raise TranspilerError(
                f"{swap_strategy} cannot implement all edges in {required_edges}."
            )
